import dataclasses

from construct import (
    Byte,
    BytesInteger,
    Const,
    Float32l,
    GreedyBytes,
    IfThenElse,
    Int8sl,
    PascalString,
    Prefixed,
    PrefixedArray,
    this,
)
from construct_typed import (
    DataclassMixin,
    DataclassStruct,
    EnumBase,
    TEnum,
    csfield,
)

from .constants import NOTE_UNKNOWN_DATA_BLOCK

Int32ul = BytesInteger(4, swapped=True)
Int32sl = BytesInteger(4, swapped=True, signed=True)


DvBytes = Prefixed(Int32ul, GreedyBytes)
DvStr = PascalString(Int32ul, "utf-8")


@dataclasses.dataclass
class DvPoint(DataclassMixin):
    x: int = csfield(Int32sl)
    y: int = csfield(Int32sl)


DvParam = Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvPoint)))


@dataclasses.dataclass
class DvTempo(DataclassMixin):
    position: int = csfield(Int32sl)
    bpm: int = csfield(Int32ul)


@dataclasses.dataclass
class DvTimeSignature(DataclassMixin):
    measure_position: int = csfield(Int32sl)
    numerator: int = csfield(Int32ul)
    denominator: int = csfield(Int32ul)


@dataclasses.dataclass
class DvNoteParameter(DataclassMixin):
    amplitude_points: list[DvPoint] = csfield(DvParam)
    frequency_points: list[DvPoint] = csfield(DvParam)
    vibrato_points: list[DvPoint] = csfield(DvParam)


@dataclasses.dataclass
class DvPhoneme(DataclassMixin):
    unknown_1: int = csfield(Int8sl)
    consonant_rate: float = csfield(Float32l)
    vowel_modified: int = csfield(Int8sl)
    medial: float = csfield(Float32l)
    rime: float = csfield(Float32l)
    ending: float = csfield(Float32l)


@dataclasses.dataclass
class DvNote(DataclassMixin):
    start: int = csfield(Int32sl)
    length: int = csfield(Int32ul)
    key: int = csfield(Int32ul)
    vibrato: int = csfield(Int32ul)
    phoneme: str = csfield(DvStr)
    word: str = csfield(DvStr)
    padding_1: int = csfield(Byte)
    note_vibrato_data: DvNoteParameter = csfield(
        Prefixed(Int32ul, DataclassStruct(DvNoteParameter))
    )
    unknown: bytes = csfield(Const(NOTE_UNKNOWN_DATA_BLOCK))
    phonemes: bytes = csfield(DataclassStruct(DvPhoneme))
    ben_depth: int = csfield(Int32ul)
    ben_length: int = csfield(Int32ul)
    por_tail: int = csfield(Int32ul)
    por_head: int = csfield(Int32ul)
    timbre: int = csfield(Int32sl)
    cross_lyric: str = csfield(DvStr)
    cross_timbre: int = csfield(Int32sl)


@dataclasses.dataclass
class DvSegment(DataclassMixin):
    start: int = csfield(Int32ul)
    length: int = csfield(Int32ul)
    name: str = csfield(DvStr)
    singer_name: str = csfield(DvStr)
    notes: list[DvNote] = csfield(
        Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvNote)))
    )
    volume_data: list[DvPoint] = csfield(DvParam)
    pitch_data: list[DvPoint] = csfield(DvParam)
    unknown_1: list[DvPoint] = csfield(DvParam)
    breath_data: list[DvPoint] = csfield(DvParam)
    gender_data: list[DvPoint] = csfield(DvParam)
    unknown_2: list[DvPoint] = csfield(DvParam)
    unknown_3: list[DvPoint] = csfield(DvParam)


@dataclasses.dataclass
class DvSingingTrack(DataclassMixin):
    name: str = csfield(DvStr)
    mute: int = csfield(Byte)
    solo: int = csfield(Byte)
    volume: int = csfield(Int32ul)
    balance: int = csfield(Int32ul)
    segments: list[DvSegment] = csfield(
        Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvSegment)))
    )


@dataclasses.dataclass
class DvAudioInfo(DataclassMixin):
    start: int = csfield(Int32ul)
    length: int = csfield(Int32ul)
    name: str = csfield(DvStr)
    path: str = csfield(DvStr)


@dataclasses.dataclass
class DvAudioTrack(DataclassMixin):
    name: str = csfield(DvStr)
    mute: int = csfield(Byte)
    solo: int = csfield(Byte)
    volume: int = csfield(Int32ul)
    balance: int = csfield(Int32ul)
    infos: list[DvAudioInfo] = csfield(
        Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvAudioInfo)))
    )


class DvTrackType(EnumBase):
    SINGING = 0
    AUDIO = 1


@dataclasses.dataclass
class DvTrack(DataclassMixin):
    track_type: DvTrackType = csfield(TEnum(Int32ul, DvTrackType))
    track_data: DvSingingTrack | DvAudioTrack = csfield(
        IfThenElse(
            this.track_type == DvTrackType.SINGING,
            DataclassStruct(DvSingingTrack),
            DataclassStruct(DvAudioTrack),
        )
    )


@dataclasses.dataclass
class DvInnerProject(DataclassMixin):
    ext_string: bytes = csfield(Const(b"ext1ext2ext3ext4ext5ext6ext7"))
    tempos: list[DvTempo] = csfield(
        Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvTempo)))
    )
    time_signatures: list[DvTimeSignature] = csfield(
        Prefixed(Int32ul, PrefixedArray(Int32ul, DataclassStruct(DvTimeSignature)))
    )
    tracks: list[DvTrack] = csfield(PrefixedArray(Int32ul, DataclassStruct(DvTrack)))


@dataclasses.dataclass
class DvProject(DataclassMixin):
    header: bytes = csfield(Const(b"SHARPKEY\x05\x00\x00\x00"))
    inner_project: DvInnerProject = csfield(Prefixed(Int32ul, DataclassStruct(DvInnerProject)))


dv_project_struct = DataclassStruct(DvProject)



import contextlib
import dataclasses
import math

import mido_fix as mido

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.search import find_index

from .model import (
    GjgjBeatItems,
    GjgjBeatStyle,
    GjgjInstrumentalTrack,
    GjgjProject,
    GjgjSingingTrack,
    GjgjTempoMap,
    GjgjTimeSignature,
    GjgjTone,
    GjgjVolumeMap,
)
from .options import InputOptions
from .singers import DEFAULT_SINGER, id2singer


@dataclasses.dataclass
class GjgjParser:
    options: InputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    ticks_in_beat: int = dataclasses.field(init=False)

    def parse_project(self, gjgj_project: GjgjProject) -> Project:
        self.ticks_in_beat = gjgj_project.tempo_map.ticks_per_quarter_note
        project = Project(
            song_tempo_list=self.parse_tempos(gjgj_project.tempo_map),
            time_signature_list=self.parse_time_signatures(gjgj_project.tempo_map.time_signature),
        )
        project.track_list = self.parse_singing_tracks(gjgj_project.tracks)
        if self.options.import_instrumental_track:
            project.track_list.extend(self.parse_instrumental_tracks(gjgj_project.accompaniments))
        return project

    def parse_tempos(self, tempo_map: GjgjTempoMap) -> list[SongTempo]:
        tempos = [
            SongTempo(
                position=tempo.time,
                bpm=mido.tempo2bpm(tempo.microseconds_per_quarter_note),
            )
            for tempo in tempo_map.tempos
        ]
        self.time_synchronizer = TimeSynchronizer(tempos)
        return tempos

    def parse_time_signatures(
        self, time_signatures: list[GjgjTimeSignature]
    ) -> list[TimeSignature]:
        if not len(time_signatures) or time_signatures[0].time != 0:
            time_signatures.insert(0, GjgjTimeSignature(time=0, numerator=4, denominator=4))
        time_signature_changes = [
            TimeSignature(
                bar_index=0,
                numerator=time_signatures[0].numerator,
                denominator=time_signatures[0].denominator,
            )
        ]
        self.first_bar_length = int(time_signature_changes[0].bar_length(self.ticks_in_beat))

        prev_ticks = 0
        measure = 0.0
        for time_signature in time_signatures[1:]:
            tick = time_signature.time
            measure += (tick - prev_ticks) / time_signature_changes[-1].bar_length(
                self.ticks_in_beat
            )
            ts_obj = TimeSignature(
                bar_index=math.floor(measure),
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            time_signature_changes.append(ts_obj)
            prev_ticks = tick
        return time_signature_changes

    def parse_singing_tracks(self, tracks: list[GjgjSingingTrack]) -> list[Track]:
        return [
            SingingTrack(
                ai_singer_name=id2singer.get(track.singer_info.display_name, DEFAULT_SINGER),
                mute=track.master_volume.mute if track.master_volume is not None else False,
                note_list=self.parse_notes(track.beat_items),
                edited_params=self.parse_params(track),
            )
            for track in tracks
        ]

    def parse_instrumental_tracks(self, accompaniments: list[GjgjInstrumentalTrack]) -> list[Track]:
        return [
            InstrumentalTrack(
                mute=accompaniment.master_volume.mute
                if accompaniment.master_volume is not None
                else False,
                audio_file_path=accompaniment.path,
                offset=round(
                    self.time_synchronizer.get_actual_ticks_from_secs(
                        accompaniment.offset / 10000000
                    )
                ),
            )
            for accompaniment in accompaniments
        ]

    def parse_notes(self, beat_items: list[GjgjBeatItems]) -> list[Note]:
        note_list = []
        for beat_item in beat_items:
            start_pos = beat_item.start_tick
            note = Note(
                start_pos=start_pos,
                length=beat_item.duration,
                lyric=beat_item.lyric.strip(),
                key_number=beat_item.track,
                pronunciation=beat_item.pinyin,
                edited_phones=self.parse_phones(beat_item, start_pos),
            )
            if beat_item.style == GjgjBeatStyle.SP:
                note.head_tag = "V"
            elif beat_item.style == GjgjBeatStyle.SIL:
                note.head_tag = "0"
            note_list.append(note)
        return note_list

    def parse_phones(self, beat_item: GjgjBeatItems, start_pos: int) -> Phones:
        phones = Phones()
        with contextlib.suppress(Exception):
            if beat_item.pre_time != 0:
                difference = round(beat_item.pre_time * 480 * 3 / 2000)
                if difference > 0:
                    phones.head_length_in_secs = self.time_synchronizer.get_actual_secs_from_ticks(
                        start_pos
                    ) - self.time_synchronizer.get_actual_secs_from_ticks(difference)
                else:
                    phones.head_length_in_secs = -1
            if beat_item.post_time != 0:
                essential_vowel_length = beat_item.duration * (2000 / 3) / 480 + beat_item.post_time
                tail_vowel_length = -beat_item.post_time
                phones.mid_ratio_over_tail = essential_vowel_length / tail_vowel_length
            else:
                phones.mid_ratio_over_tail = -1
        return phones

    def parse_params(self, track: GjgjSingingTrack) -> Params:
        params = Params()
        if self.options.import_pitch:
            params.pitch = self.parse_pitch_curve(track.tone)
        if self.options.import_volume:
            params.volume = self.parse_volume_curve(track.volume_map)
        return params

    def pitch_time_to_position(self, pitch_time: float) -> int:
        return round(pitch_time * 5) + self.first_bar_length

    def parse_pitch_curve(self, tone: GjgjTone) -> ParamCurve:
        pitch_curve = ParamCurve()
        pitch_curve.points.append(Point.start_point())
        with contextlib.suppress(Exception):
            for mod_range in tone.modify_ranges:
                left_point = Point(self.pitch_time_to_position(mod_range.x), -100)
                right_point = Point(self.pitch_time_to_position(mod_range.y), -100)
                index = find_index(
                    tone.modifies,
                    lambda p: (mod_range.x <= p.time <= mod_range.y),
                )
                if index == -1:
                    continue
                pitch_curve.points.append(left_point)
                while index < len(tone.modifies) and tone.modifies[index].x <= mod_range.y:
                    pitch_curve.points.append(
                        Point(
                            self.pitch_time_to_position(tone.modifies[index].x),
                            self.pitch_time_to_position(tone.modifies[index].y),
                        )
                    )
                    index += 1
                pitch_curve.points.append(right_point)
        pitch_curve.points.append(Point.end_point())
        return pitch_curve

    def parse_volume_curve(self, volume_map: list[GjgjVolumeMap]) -> ParamCurve:
        volume_curve = ParamCurve()
        volume_curve.points.root = [
            Point(round(volume_item.time), round(volume_item.volume)) for volume_item in volume_map
        ]
        return volume_curve



from pydantic import BaseModel

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
    StaticTempoMixin,
)


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
    BaseModel,
):
    pass


class OutputOptions(StaticTempoMixin, BaseModel):
    pass



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VoiSonaMobileProject, model_to_value_tree
from .options import InputOptions, OutputOptions
from .value_tree import JUCENode, build_tree_dict
from .voisona_mobile_generator import VoiSonaMobileGenerator
from .voisona_mobile_parser import VoiSonaMobileParser


class VoiSonaMobileConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        value_tree = JUCENode.parse(path.read_bytes())
        tree_dict = build_tree_dict(value_tree)
        tsmsln_project = VoiSonaMobileProject.model_validate(tree_dict["MobileSongEditor"])
        return VoiSonaMobileParser(options).parse_project(tsmsln_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        tsmsln_project = VoiSonaMobileGenerator(options).generate_project(project)
        value_tree = model_to_value_tree(tsmsln_project, name="MobileSongEditor")
        path.write_bytes(JUCENode.build(value_tree))



from __future__ import annotations

import bisect
import functools
import math
from types import SimpleNamespace
from typing import Annotated, Literal

from pydantic import Field, computed_field, field_validator

from libresvip.core.constants import DEFAULT_BPM, TICKS_IN_BEAT
from libresvip.model.base import BaseModel
from libresvip.utils.audio import audio_path_validator
from libresvip.utils.music_math import linear_interpolation

ParamType = SimpleNamespace(
    CURVE="Curve",
    NUMERICAL="Numerical",
    OPTIONS="Options",
)


class BaseExpression(BaseModel):
    name: str
    abbr: str
    min_: float = Field(alias="min")
    max_: float = Field(alias="max")
    default_value: float
    is_flag: bool


class CurveExpression(BaseExpression):
    type_: Literal["Curve"] = Field(ParamType.CURVE, alias="type")
    flag: str


class NumericalExpression(BaseExpression):
    type_: Literal["Numerical"] = Field(ParamType.NUMERICAL, alias="type")
    flag: str


class OptionsExpression(BaseExpression):
    type_: Literal["Options"] = Field(ParamType.OPTIONS, alias="type")
    options: list[str] = Field(default_factory=list)


UExpressionDescriptor = Annotated[
    CurveExpression | NumericalExpression | OptionsExpression,
    Field(discriminator="type_"),
]


class UCurve(BaseModel):
    xs: list[int] = Field(default_factory=list)
    ys: list[int] = Field(default_factory=list)
    abbr: str | None = None

    @property
    def is_empty(self) -> bool:
        return len(self.xs) == 0 or all(y == 0 for y in self.ys)

    def sample(self, x: int) -> int:
        idx = bisect.bisect_left(self.xs, x)
        if idx < len(self.xs) and self.xs[idx] == x:
            return self.ys[idx]
        elif 0 < idx < len(self.xs):
            return round(
                linear_interpolation(
                    x,
                    (self.xs[idx - 1], self.ys[idx - 1]),
                    (self.xs[idx], self.ys[idx]),
                )
            )
        return 0


class PitchPoint(BaseModel):
    x: float
    y: float
    shape: Literal["io", "l", "i", "o"] | None = "io"


class UTempo(BaseModel):
    position: int = 0
    bpm: float = DEFAULT_BPM


class UTimeSignature(BaseModel):
    bar_position: int = 0
    beat_per_bar: int = 4
    beat_unit: int = 4


class URendererSettings(BaseModel):
    renderer: str | None = None
    resampler: str | None = None
    wavtool: str | None = None


class UTrack(BaseModel):
    track_name: str | None = None
    track_color: str | None = None
    track_expressions: list[UExpression] | None = Field(default_factory=list)
    voice_color_names: list[str] | None = None
    singer: str | None = None
    phonemizer: str | None = None
    renderer_settings: URendererSettings | None = None
    mute: bool = False
    solo: bool = False
    volume: float = 0.0
    pan: float | None = None


class UPitch(BaseModel):
    data: list[PitchPoint] = Field(default_factory=list)
    snap_first: bool | None = None


class UVibrato(BaseModel):
    length: float = 75.0
    period: float = 175.0
    depth: float = 25.0
    in_value: float = Field(10.0, alias="in")
    out: float = 10.0
    shift: float = 0.0
    drift: float = 0.0
    vol_link: float = 0.0

    @functools.cached_property
    def normalized_start(self) -> float:
        return 1.0 - self.length / 100.0

    def evaluate(self, n_pos: float, n_period: float, note: UNote) -> tuple[float, float]:
        n_start = self.normalized_start
        n_in = self.length / 100.0 * self.in_value / 100.0
        n_in_pos = n_start + n_in
        n_out = self.length / 100.0 * self.out / 100.0
        n_out_pos = 1.0 - n_out
        t = (n_pos - n_start) / n_period + self.shift / 100.0
        y = math.sin(math.tau * t) * self.depth
        if n_pos < n_start:
            y = 0.0
        elif n_pos < n_in_pos:
            y *= (n_pos - n_start) / n_in
        elif n_pos > n_out_pos:
            y *= (1.0 - n_pos) / n_out
        return (note.position + note.duration * n_pos, note.tone + y / 100.0)


class UExpression(BaseModel):
    index: int | None = None
    abbr: str
    value: float


class UPhonemeOverride(BaseModel):
    index: int
    phoneme: str | None = None
    offset: int | None = None
    preutter_delta: float | None = None
    overlap_delta: float | None = None


class UNote(BaseModel):
    position: int
    duration: int
    tone: int
    lyric: str
    pitch: UPitch = Field(default_factory=UPitch)
    vibrato: UVibrato = Field(default_factory=UVibrato)
    note_expressions: list[UExpression] | None = Field(default_factory=list)  # deprecated
    phoneme_expressions: list[UExpression] | None = Field(default_factory=list)
    phoneme_overrides: list[UPhonemeOverride] | None = Field(default_factory=list)

    @functools.cached_property
    def end(self) -> int:
        return self.position + self.duration


class UPart(BaseModel):
    name: str
    position: int = 0
    track_no: int
    comment: str | None = ""


class UVoicePart(UPart):
    notes: list[UNote] = Field(default_factory=list)
    curves: list[UCurve] = Field(default_factory=list)

    @computed_field
    def duration(self) -> int:
        return max(self.notes[-1].end - self.position if self.notes else 0, 480)


class UWavePart(UPart):
    relative_path: str
    file_duration_ms: float = 0
    skip_ms: float = 0
    trim_ms: float = 0

    validate_relative_path = field_validator("relative_path", mode="before")(audio_path_validator)


class USTXProject(BaseModel):
    name: str = "New Project"
    comment: str = ""
    output_dir: str = "Vocal"
    cache_dir: str = "UCache"
    ustx_version: str | float = "0.6"
    bpm: float = DEFAULT_BPM
    beat_per_bar: int = 4
    beat_unit: int = 4
    resolution: int = TICKS_IN_BEAT
    time_signatures: list[UTimeSignature] = Field(default_factory=list)
    tempos: list[UTempo] = Field(default_factory=list)
    expressions: dict[str, UExpressionDescriptor] | None = None
    tracks: list[UTrack] = Field(default_factory=list)
    voice_parts: list[UVoicePart] = Field(default_factory=list)
    wave_parts: list[UWavePart] = Field(default_factory=list)
    exp_selectors: list[str] | None = None
    exp_primary: int | None = 0
    exp_secondary: int | None = 1
    key: int | None = 0



# mypy: disable-error-code="misc"
import enum
from typing import Annotated

from pydantic import Field

from libresvip.utils.translation import gettext_lazy as _


class VsqxVersion(enum.IntEnum):
    VSQ3: Annotated[int, Field(title=_("VSQx 3"))] = 3
    VSQ4: Annotated[int, Field(title=_("VSQx 4"))] = 4


class VocaloidLanguage(enum.IntEnum):
    JAPANESE: Annotated[int, Field(title=_("日本語"))] = 0
    ENGLISH: Annotated[int, Field(title=_("English"))] = 1
    KOREAN: Annotated[int, Field(title=_("한국어"))] = 2
    SPANISH: Annotated[int, Field(title=_("Español"))] = 3
    SIMPLIFIED_CHINESE: Annotated[int, Field(title=_("简体中文"))] = 4



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

import portion

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import ParamCurve, Points
from libresvip.model.point import Point

if TYPE_CHECKING:
    from libresvip.model.pitch_simulator import PitchSimulator


@dataclasses.dataclass
class RelativePitchCurve:
    first_bar_length: int = TICKS_IN_BEAT * 4
    lower_bound: float = 0.0
    upper_bound: float = portion.inf
    pitch_interval: int = 5

    def to_absolute(self, points: list[Point], pitch_simulator: PitchSimulator) -> ParamCurve:
        return ParamCurve(
            points=Points(root=self._convert_relativity(points, pitch_simulator, to_absolute=True))
        )

    def _convert_relativity(
        self,
        points: list[Point],
        pitch_simulator: PitchSimulator,
        to_absolute: bool = False,
    ) -> list[Point]:
        converted_data: list[Point] = []
        prev_x = None
        prev_y: float | None = None
        for point in points:
            pos = point.x + (0 if to_absolute else -self.first_bar_length)
            cur_x = point.x + (self.first_bar_length if to_absolute else -self.first_bar_length)
            if (base_key := pitch_simulator.pitch_at_ticks(pos)) is None:
                y = None
            elif not to_absolute and point.y == -100:
                if point.x in [-192000, 1073741823]:
                    continue
                y = 0.0
            else:
                y = point.y + (base_key if to_absolute else -base_key)
            if (
                y is not None
                and prev_x is not None
                and prev_y is not None
                and converted_data
                and cur_x - prev_x > self.pitch_interval
            ):
                for tick in range(prev_x + self.pitch_interval, cur_x, self.pitch_interval):
                    tick_pos = tick + (-self.first_bar_length if to_absolute else 0)
                    if (tick_key := pitch_simulator.pitch_at_ticks(tick_pos)) is not None:
                        if to_absolute:
                            converted_data.append(Point(x=tick, y=round(prev_y + tick_key)))
                        else:
                            converted_data.append(
                                Point(
                                    x=tick,
                                    y=round(
                                        prev_y + (y - prev_y) * (tick - prev_x) / (cur_x - prev_x)
                                    ),
                                )
                            )
            if y is not None:
                if to_absolute and prev_y is None:
                    converted_data.append(Point(x=cur_x, y=-100))
                cur_point = Point(x=cur_x, y=round(y))
                converted_data.append(cur_point)
                if to_absolute:
                    if point.y:
                        prev_y = point.y
                    else:
                        converted_data.append(Point(x=cur_x, y=-100))
                        prev_y = None
                else:
                    prev_y = None if point.y == -100 else y
            else:
                prev_y = None
            prev_x = cur_x
        if converted_data and to_absolute:
            if self.lower_bound == 0:
                converted_data.insert(0, Point.start_point())
            if self.upper_bound == portion.inf:
                converted_data.extend((converted_data[-1]._replace(y=-100), Point.end_point()))
        return converted_data

    def from_absolute(
        self,
        points: list[Point],
        pitch_simulator: PitchSimulator,
    ) -> list[Point]:
        return self._convert_relativity(points, pitch_simulator, to_absolute=False)



import dataclasses

from pysubs2 import SSAEvent, SSAFile

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, Project, SingingTrack
from libresvip.utils.text import LATIN_ALPHABET, SYMBOL_PATTERN

from .options import OutputOptions, SplitOption


@dataclasses.dataclass
class SrtGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> SSAFile:
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        if self.options.track_index == -1:
            singing_track = next(
                track for track in project.track_list if isinstance(track, SingingTrack)
            )
        else:
            singing_track = project.track_list[self.options.track_index]
        note_list = singing_track.note_list
        buffer = []
        lyric_lines = SSAFile()
        for i, note in enumerate(note_list):
            buffer.append(note)
            commit_flag = False
            condition_symbol = SYMBOL_PATTERN.search(note.lyric) is not None
            condition_gap = (
                i + 1 < len(note_list) and note_list[i + 1].start_pos - note.end_pos >= 60
            )
            if self.options.split_by == SplitOption.SYMBOL:
                commit_flag = condition_symbol
            elif self.options.split_by == SplitOption.GAP:
                commit_flag = condition_gap
            elif self.options.split_by == SplitOption.BOTH:
                commit_flag = condition_symbol or condition_gap
            if i + 1 == len(note_list):
                commit_flag = True
            if commit_flag:
                self.commit_current_lyric_line(lyric_lines, buffer)
                buffer = []
        return lyric_lines

    def lyric_included(self, lyric: str) -> bool:
        if self.options.ignore_slur_notes:
            return lyric != "-"
        else:
            return True

    def commit_current_lyric_line(self, lyric_lines: SSAFile, buffer: list[Note]) -> None:
        start_time = int(self.synchronizer.get_actual_secs_from_ticks(buffer[0].start_pos) * 1000)
        end_time = int(self.synchronizer.get_actual_secs_from_ticks(buffer[-1].end_pos) * 1000)
        lyrics = "".join(
            SYMBOL_PATTERN.sub("", note.lyric)
            + (" " if LATIN_ALPHABET.search(note.lyric) is not None else "")
            for note in buffer
            if self.lyric_included(note.lyric)
        ).strip()
        lyric_lines.append(
            SSAEvent(
                start=start_time,
                end=end_time,
                text=lyrics,
            )
        )



import functools

import more_itertools
import portion

from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.model.point import Point
from libresvip.utils.music_math import cosine_easing_in_out_interpolation

from .model import PpsfDvlTrackEvent, PpsfNote


def ppsf_key_interval_dict(
    event_list: list[PpsfDvlTrackEvent], note_list: list[PpsfNote]
) -> PiecewiseIntervalDict:
    interval_dict = PiecewiseIntervalDict()
    if len(event_list) == 1:
        interval_dict[portion.closedopen(0, portion.inf)] = event_list[0].note_number
    else:
        for (
            is_first,
            is_last,
            ((prev_note, prev_event), (next_note, next_event)),
        ) in more_itertools.mark_ends(more_itertools.pairwise(zip(event_list, note_list))):
            if is_first:
                interval_dict[
                    portion.closedopen(
                        0,
                        prev_note.pos + prev_event.portamento_offset + prev_event.portamento_length,
                    )
                ] = prev_note.note_number
            if next_event.portamento_length:
                if (
                    prev_portamento_end := prev_note.pos
                    + prev_event.portamento_offset
                    + prev_event.portamento_length
                ) < (next_portamento_start := next_note.pos + next_event.portamento_offset):
                    interval_dict[
                        portion.closedopen(
                            prev_portamento_end,
                            next_portamento_start,
                        )
                    ] = prev_note.note_number
                interval_dict[
                    portion.closedopen(
                        next_portamento_start,
                        next_note.pos + next_event.portamento_offset + next_event.portamento_length,
                    )
                ] = functools.partial(
                    cosine_easing_in_out_interpolation,  # type: ignore[call-arg]
                    start=Point(
                        x=next_note.pos + next_event.portamento_offset,
                        y=prev_note.note_number,
                    ),
                    end=Point(
                        x=next_note.pos
                        + next_event.portamento_offset
                        + next_event.portamento_length,
                        y=next_note.note_number,
                    ),
                )
            elif (
                prev_portamento_end := prev_note.pos
                + prev_event.portamento_offset
                + prev_event.portamento_length
            ) < prev_note.end_pos:
                interval_dict[
                    portion.closedopen(
                        prev_portamento_end,
                        prev_note.end_pos,
                    )
                ] = prev_note.note_number
            if is_last:
                interval_dict[
                    portion.closedopen(
                        next_note.pos + next_event.portamento_offset + next_event.portamento_length,
                        portion.inf,
                    )
                ] = next_note.note_number
    return interval_dict



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_extra_types.color import Color

from libresvip.model.option_mixins import SelectSingleTrackMixin
from libresvip.utils.translation import gettext_lazy as _


class TextAlignOption(Enum):
    START: Annotated[str, Field(title=_("Align to left"))] = "start"
    MIDDLE: Annotated[str, Field(title=_("Align to middle"))] = "middle"
    END: Annotated[str, Field(title=_("Align to right"))] = "end"


class TextPositionOption(Enum):
    UPPER: Annotated[str, Field(title=_("Above the note"))] = "upper"
    INNER: Annotated[str, Field(title=_("Inside the note"))] = "inner"
    LOWER: Annotated[str, Field(title=_("Below the note"))] = "lower"
    NONE: Annotated[str, Field(title=_("Don't show"))] = "none"


class OutputOptions(SelectSingleTrackMixin, BaseModel):
    pixel_per_beat: int = Field(48, title=_("Pixels per beat"), description=_("Unit: pixel."))
    note_height: int = Field(24, title=_("Note height"), description=_("Unit: pixel."))
    note_round: int = Field(4, title=_("Note radius"), description="Unit: pixel.")
    note_fill_color: Color = Field(
        "#CCFFCC",
        title=_("Note fill color"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    note_stroke_color: Color = Field(
        "#006600",
        title=_("Note stroke color"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    note_stroke_width: int = Field(1, title=_("Note stroke width"), description="Unit: pixel.")
    pitch_stroke_color: Color = Field(
        "#99aa99",
        title=_("Pitch stroke color"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    pitch_stroke_width: int = Field(2, title=_("Pitch stroke width"), description=_("Unit: pixel."))
    lyric_position: TextPositionOption = Field(
        TextPositionOption.LOWER,
        title=_("Lyric position"),
    )
    pronounciation_position: TextPositionOption = Field(
        TextPositionOption.INNER,
        title=_("Pronounciation position"),
    )
    text_align: TextAlignOption = Field(
        TextAlignOption.START,
        title=_("Text align"),
    )
    inner_text_color: Color = Field(
        "#000000",
        title=_("Inner text color"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    side_text_color: Color = Field(
        "#000000",
        title=_("The color of the text on the top and bottom of the note"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    show_grid: bool = Field(default=False, title=_("Show grid lines"))
    grid_color: Color = Field(
        "#CCCCCC",
        title=_("Grid line color"),
        description=_("CSS color value, e.g. #FF0000, #66CCFF, rgba(255,0,0,0.5), etc."),
    )
    grid_stroke_width: int = Field(1, title=_("Grid line width"), description=_("Unit: pixel."))



import dataclasses
import math

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import Note, SingingTrack
from libresvip.model.point import Point

from .model import NotePositionParameters
from .options import OutputOptions, TextAlignOption

PADDING: int = 4


@dataclasses.dataclass
class CoordinateHelper:
    options: OutputOptions
    pitch_position_offset: int
    position_range_start: int = dataclasses.field(init=False)
    position_range_end: int = dataclasses.field(init=False)
    key_range_start: int = dataclasses.field(init=False)
    key_range_end: int = dataclasses.field(init=False)

    def calculate_range(self, track: SingingTrack) -> None:
        if not len(track.edited_params.pitch.points):
            track.edited_params.pitch.points.root = [Point.start_point(), Point.end_point()]
        self.position_range_start = min(
            track.note_list[0].start_pos, track.edited_params.pitch.points[1].x
        )
        self.position_range_end = max(
            track.note_list[-1].end_pos, track.edited_params.pitch.points[-2].x
        )
        self.key_range_start = min(note.key_number for note in track.note_list)
        self.key_range_end = max(note.key_number for note in track.note_list)
        for point in track.edited_params.pitch.points.root[1:-1]:
            if point.y == -100:
                continue
            self.key_range_start = min(self.key_range_start, math.floor(point.y / 100))
            self.key_range_end = max(self.key_range_end, math.ceil(point.y / 100))

    def get_note_position_parameters(self, note: Note) -> NotePositionParameters:
        text_x = 0
        if self.options.text_align == TextAlignOption.START:
            text_x = int(
                (note.start_pos - self.position_range_start)
                * self.options.pixel_per_beat
                / TICKS_IN_BEAT
                + PADDING
            )
        elif self.options.text_align == TextAlignOption.MIDDLE:
            text_x = int(
                ((note.start_pos + note.end_pos) / 2 - self.position_range_start)
                * self.options.pixel_per_beat
                / TICKS_IN_BEAT
            )
        elif self.options.text_align == TextAlignOption.END:
            text_x = int(
                (note.end_pos - self.position_range_start)
                * self.options.pixel_per_beat
                / TICKS_IN_BEAT
                - PADDING
            )
        return NotePositionParameters(
            point_1=(
                int(
                    (note.start_pos - self.position_range_start)
                    * self.options.pixel_per_beat
                    / TICKS_IN_BEAT
                ),
                int((self.key_range_end - note.key_number) * self.options.note_height),
            ),
            point_2=(
                int(
                    (note.end_pos - self.position_range_start)
                    * self.options.pixel_per_beat
                    / TICKS_IN_BEAT
                ),
                int((self.key_range_end - note.key_number + 1) * self.options.note_height),
            ),
            text_size=self.font_size,
            inner_text=(
                text_x,
                int(
                    (self.key_range_end - note.key_number + 1) * self.options.note_height
                    - PADDING * 1.5
                ),
            ),
            upper_text=(
                text_x,
                int((self.key_range_end - note.key_number) * self.options.note_height - PADDING),
            ),
            lower_text=(
                text_x,
                (self.key_range_end - note.key_number + 2) * self.options.note_height - PADDING,
            ),
        )

    def get_pitch_point(self, param_point: Point) -> tuple[float, float]:
        return (
            (param_point.x - self.position_range_start - self.pitch_position_offset)
            * self.options.pixel_per_beat
            / TICKS_IN_BEAT,
            (self.key_range_end - param_point.y / 100 + 0.5) * self.options.note_height,
        )

    @property
    def size(self) -> tuple[float, float]:
        return (
            (self.position_range_end - self.position_range_start)
            * self.options.pixel_per_beat
            / TICKS_IN_BEAT,
            (self.key_range_end - self.key_range_start + 1) * self.options.note_height,
        )

    @property
    def font_size(self) -> int:
        return self.options.note_height - 2 * PADDING

    @property
    def text_anchor(self) -> str:
        return self.options.text_align.value



# mypy: disable-error-code="attr-defined"
import ctypes
from ctypes.wintypes import MSG
from typing import SupportsInt

from PySide6.QtCore import QByteArray, QPoint, QRect, Qt, Slot
from PySide6.QtGui import QGuiApplication, QMouseEvent, QWindow
from PySide6.QtQml import QmlElement
from PySide6.QtQuick import QQuickItem, QQuickWindow

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from . import win32_constants as win32con

QML_IMPORT_NAME = "FramelessWindow"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


class MARGINS(ctypes.Structure):
    _fields_ = (
        ("cxLeftWidth", ctypes.c_int),
        ("cxRightWidth", ctypes.c_int),
        ("cyTopHeight", ctypes.c_int),
        ("cyBottomHeight", ctypes.c_int),
    )


@QmlElement
class FramelessWindow(QQuickWindow):
    def __init__(self, parent: QWindow | None = None, border_width: int = 5) -> None:
        super().__init__(parent)
        self.flags: Qt.WindowType = (
            self.flags
            | Qt.WindowType.FramelessWindowHint
            | Qt.WindowType.Window
            | Qt.WindowType.WindowMinMaxButtonsHint
        )
        self.border_width = border_width
        self.maximize_btn: QQuickItem | None = None
        self.maximize_btn_hovered = False
        self.set_borderless()
        screen_geometry = self.screen().available_geometry
        self.set_position(
            (screen_geometry.width() - 1200) // 2,
            (screen_geometry.height() - 800) // 2,
        )
        self.prev_visibility = None

    @Slot()
    def click_maximize_btn(self) -> None:
        if self.visibility != QQuickWindow.Visibility.Maximized:
            ctypes.windll.user32.SendMessageW(
                self.hwnd, win32con.WM_SYSCOMMAND, win32con.SC_MAXIMIZE, 0
            )
        else:
            ctypes.windll.user32.SendMessageW(
                self.hwnd, win32con.WM_SYSCOMMAND, win32con.SC_RESTORE, 0
            )

    def get_point_from_lparam(self, l_param: int) -> tuple[int, int]:
        pixel_ratio = self.screen().device_pixel_ratio
        return (
            (ctypes.c_short(l_param & 0xFFFF).value) // pixel_ratio - self.x,
            (ctypes.c_short((l_param >> 16) & 0xFFFF).value) // pixel_ratio - self.y,
        )

    @property
    def hwnd(self) -> int:
        return self.win_id()

    @property
    def is_composition_enabled(self) -> bool:
        b_result = ctypes.c_int(0)
        ctypes.windll.dwmapi.DwmIsCompositionEnabled(ctypes.byref(b_result))
        return bool(b_result.value)

    def set_borderless(self) -> None:
        hwnd = self.hwnd
        user32 = ctypes.windll.user32

        style = user32.GetWindowLongW(hwnd, win32con.GWL_STYLE)

        user32.SetWindowLongW(
            hwnd,
            win32con.GWL_STYLE,
            style
            | win32con.WS_MAXIMIZEBOX
            | win32con.WS_CAPTION
            | win32con.CS_DBLCLKS
            | win32con.WS_THICKFRAME,
        )

        style = user32.GetWindowLongW(hwnd, win32con.GWL_EXSTYLE)
        style &= ~win32con.WS_EX_LAYERED
        user32.SetWindowLongW(hwnd, win32con.GWL_EXSTYLE, style)

    def add_dwm_effect(self) -> SupportsInt | None:
        if not self.is_composition_enabled:
            return None

        dwmapi = ctypes.windll.dwmapi

        margins = MARGINS(-1, -1, -1, -1)
        return dwmapi.DwmExtendFrameIntoClientArea(self.hwnd, ctypes.byref(margins))

    def native_event(
        self, event_type: QByteArray | bytes | bytearray | memoryview, message: SupportsInt
    ) -> object:
        if self.maximize_btn is None:
            self.maximize_btn = self.find_child(QQuickItem, "maximizeButton")
        if event_type == b"windows_generic_MSG":
            msg = MSG.from_address(int(message))

            if msg.message == win32con.WM_NCHITTEST and self.border_width is not None:
                if msg.hWnd == self.hwnd:
                    x_pos, y_pos = self.get_point_from_lparam(msg.lParam)
                    bw = (
                        0
                        if self.visibility == QQuickWindow.Visibility.Maximized
                        else self.border_width
                    )
                    lx = x_pos < bw
                    rx = x_pos > self.width - bw
                    ty = y_pos < bw
                    by = y_pos > self.height - bw
                    if lx and ty:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTTOPLEFT
                    elif rx and by:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTBOTTOMRIGHT
                    elif rx and ty:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTTOPRIGHT
                    elif lx and by:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTBOTTOMLEFT
                    elif ty:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTTOP
                    elif by:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTBOTTOM
                    elif lx:
                        if self.visibility != QQuickWindow.Visibility.Maximized:
                            return True, win32con.HTLEFT
                    elif rx and self.visibility != QQuickWindow.Visibility.Maximized:
                        return True, win32con.HTRIGHT
                    if self.maximize_btn is not None:
                        top_left = self.maximize_btn.map_to_global(QPoint(0, 0))
                        rect = QRect(
                            top_left.x() - self.x,
                            top_left.y() - self.y,
                            self.maximize_btn.width,
                            self.maximize_btn.height,
                        )
                        if rect.contains(x_pos, y_pos):
                            return True, win32con.HTMAXBUTTON
            elif msg.message in [
                win32con.WM_NCMOUSEHOVER,
                win32con.WM_NCMOUSEMOVE,
                win32con.WM_NCMOUSELEAVE,
                win32con.WM_NCLBUTTONDOWN,
                win32con.WM_NCLBUTTONUP,
            ]:
                if self.maximize_btn is not None:
                    self.handle_mouse_event(msg)
            elif msg.message == win32con.WM_NCCALCSIZE:
                return True, win32con.WVR_REDRAW if msg.wParam else 0
            elif msg.message == win32con.WM_ACTIVATE:
                if (hr := self.add_dwm_effect()) is not None:
                    return True, hr
            elif msg.message == win32con.WM_SYSCOMMAND:
                if (
                    msg.wParam == win32con.SC_RESTORE
                    and self.visibility == QQuickWindow.Visibility.Minimized
                ):
                    if self.prev_visibility == QQuickWindow.Visibility.Maximized:
                        self.show_maximized()
                    else:
                        self.show_normal()
                    return True, 0
            elif msg.message == win32con.WM_SIZE and msg.wParam == win32con.SIZE_MINIMIZED:
                self.prev_visibility = self.visibility
        return super().native_event(event_type, message)

    def handle_mouse_event(self, msg: MSG) -> None:
        if self.maximize_btn is None:
            return
        x_pos, y_pos = self.get_point_from_lparam(msg.lParam)
        top_left = self.maximize_btn.map_to_global(QPoint(0, 0))
        rect = QRect(
            top_left.x() - self.x,
            top_left.y() - self.y,
            self.maximize_btn.width,
            self.maximize_btn.height,
        )

        maximize_btn_hovered = rect.contains(x_pos, y_pos)
        if maximize_btn_hovered:
            if msg.message == win32con.WM_NCLBUTTONDOWN:
                mouse_event_type = QMouseEvent.Type.MouseButtonPress
            elif msg.message == win32con.WM_NCLBUTTONUP:
                mouse_event_type = QMouseEvent.Type.MouseButtonRelease
            mouse_event_type = (
                QMouseEvent.Type.HoverMove
                if self.maximize_btn_hovered
                else QMouseEvent.Type.HoverEnter
            )
        else:
            mouse_event_type = QMouseEvent.Type.HoverLeave
        mouse_btn = (
            Qt.MouseButton.LeftButton
            if msg.message in [win32con.WM_NCLBUTTONDOWN, win32con.WM_NCLBUTTONUP]
            else Qt.MouseButton.NoButton
        )
        QGuiApplication.send_event(
            self.maximize_btn,
            QMouseEvent(
                mouse_event_type,
                QPoint(),
                mouse_btn,
                mouse_btn,
                Qt.KeyboardModifier.NoModifier,
            ),
        )
        self.maximize_btn_hovered = maximize_btn_hovered



# Ported from QNrbf by SineStriker
# mypy: disable-error-code="attr-defined"
import dataclasses
import enum
import inspect
import math
import pathlib
from collections import defaultdict
from queue import Queue
from typing import Any, Union, cast, get_args, get_origin

from construct import Container, ListContainer

from .binary_models import (
    BinaryArrayTypeEnum,
    BinaryTypeEnum,
    PrimitiveTypeEnum,
    RecordTypeEnum,
    SVIPFile,
    classes_by_id,
    libraries_by_id,
    references_by_id,
)
from .constants import (
    LIBRARY_NAME_SINGING_TOOL_LIBRARY,
    LIBRARY_NAME_SINGING_TOOL_MODEL,
)
from .nrbf_iobase import NrbfIOBase
from .xstudio_models import (
    XSAppModel,
    XSBuf,
    XSBufList,
)


def class_def_factory() -> dict[str, dict[str, Any]]:
    return defaultdict(dict)


@dataclasses.dataclass
class SvipWriter(NrbfIOBase):
    ids: Queue[int] = dataclasses.field(default_factory=Queue)
    id_max: int = dataclasses.field(default=0)
    model_library_id: int = dataclasses.field(default=0)
    lib_library_id: int = dataclasses.field(default=0)
    written_ids: set[int] = dataclasses.field(default_factory=set)
    class_defs: dict[str, dict[str, Any]] = dataclasses.field(default_factory=class_def_factory)
    svip_file: dict[str, Any] = dataclasses.field(init=False)

    def enq(self) -> int:
        self.id_max += 1
        self.ids.put(self.id_max)
        return self.id_max

    def deq(self) -> int:
        return self.ids.get()

    def write(self, path: pathlib.Path, version: str, model: XSAppModel) -> None:
        self.svip_file = {
            "magic": version[:4],
            "version": version[4:],
            "record_stream": [],
        }
        self.id_max += 1
        app_model_id = self.id_max
        self.header = {
            "root_id": app_model_id,
            "header_id": -1,
            "major_version": 1,
            "minor_version": 0,
        }
        self.svip_file["record_stream"].append(
            {
                "record_type_enum": RecordTypeEnum.SerializedStreamHeader,
                "obj": self.header,
            }
        )
        self.model_library_id = self.write_library(LIBRARY_NAME_SINGING_TOOL_MODEL)
        self.lib_library_id = self.write_library(LIBRARY_NAME_SINGING_TOOL_LIBRARY)
        self.svip_file["record_stream"].append(self.write_dataclass(model, app_model_id))
        while not self.ids.empty():
            not_written = self.deq()
            ref = references_by_id[self.cur_thread_id][not_written]
            if dataclasses.is_dataclass(ref["real_obj"]):
                self.svip_file["record_stream"].append(
                    self.write_dataclass(ref["real_obj"], not_written, ref["subcon_class_name"])
                )
            else:
                self.svip_file["record_stream"].append(ref["real_obj"])

        self.svip_file["record_stream"].append(
            {
                "record_type_enum": RecordTypeEnum.MessageEnd,
                "obj": {},
            }
        )
        path.write_bytes(SVIPFile.build(self.svip_file))

    def write_library(self, library_name: str) -> int:
        self.id_max += 1
        result = self.id_max
        model_library = {"library_id": result, "library_name": library_name}
        libraries_by_id[self.cur_thread_id][
            model_library["library_id"]  # type: ignore[index]
        ] = model_library["library_name"]  # type: ignore[assignment]
        self.svip_file["record_stream"].append(
            {
                "record_type_enum": RecordTypeEnum.BinaryLibrary,
                "obj": model_library,
            }
        )
        return result

    def create_string(self, value: str) -> dict[str, Any]:
        self.id_max += 1
        return {
            "record_type_enum": RecordTypeEnum.BinaryObjectString,
            "obj": {
                "object_id": self.id_max,
                "value": value,
            },
        }

    def create_reference(
        self,
        value: Container,
        object_id: int,
        subcon_class_name: str | None,
    ) -> dict[str, Any]:
        result = {
            "record_type_enum": RecordTypeEnum.MemberReference,
            "obj": {
                "id_ref": object_id,
            },
        }
        if subcon_class_name is not None and "`1" in subcon_class_name:
            subcon_class_name = subcon_class_name.split("[[", 1)[-1].split(", ", 1)[0]
        if object_id not in references_by_id[self.cur_thread_id]:
            references_by_id[self.cur_thread_id][object_id] = {
                "id_ref": object_id,
                "subcon_class_name": subcon_class_name,
                "real_obj": value,
            }
        return result

    def write_null_array(self, value: int) -> dict[str, Any]:
        if value == 1:
            return {
                "record_type_enum": RecordTypeEnum.ObjectNull,
                "obj": {},
            }
        elif value <= 256:
            return {
                "record_type_enum": RecordTypeEnum.ObjectNullMultiple256,
                "obj": {
                    "null_count": value,
                },
            }
        else:
            return {
                "record_type_enum": RecordTypeEnum.ObjectNullMultiple,
                "obj": {
                    "null_count": value,
                },
            }

    def write_binary_array(
        self, values: list[bytes], type_name: str, library_id: int
    ) -> dict[str, Any]:
        object_id = self.enq()
        padded_length = max(4, 2 ** math.ceil(math.log2(len(values))) if values else 0)
        result = {
            "record_type_enum": RecordTypeEnum.BinaryArray,
            "obj": {
                "object_id": object_id,
                "binary_array_type_enum": BinaryArrayTypeEnum.Single,
                "rank": 1,
                "lengths": [padded_length],
                "lower_bounds": None,
                "binary_type_enum": BinaryTypeEnum.Class,
                "info": {
                    "type_name": type_name.replace("InstrumentTrack", "ITrack").replace(
                        "SingingTrack", "ITrack"
                    ),
                    "library_id": library_id,
                },
                "member_values": [
                    {
                        "total": padded_length,
                        "real_obj": self.create_reference(value, self.enq(), type_name),
                    }
                    for value in values
                ],
            },
        }
        if len(values) < padded_length:
            result["obj"]["member_values"].append(
                {
                    "total": padded_length,
                    "real_obj": self.write_null_array(padded_length - len(values)),
                }
            )
        references_by_id[self.cur_thread_id][object_id] = {
            "id_ref": object_id,
            "real_obj": result,
        }
        return self.create_reference(result, object_id, None)

    def create_primitive_array(self, values: ListContainer) -> dict[str, Any]:
        object_id = self.enq()
        result = {
            "record_type_enum": RecordTypeEnum.ArraySinglePrimitive,
            "obj": {
                "array_info": {
                    "object_id": object_id,
                    "length": len(values),
                },
                "primitive_type_enum": PrimitiveTypeEnum.Byte,
                "member_values": values,
            },
        }
        references_by_id[self.cur_thread_id][object_id] = {
            "id_ref": object_id,
            "real_obj": result,
        }
        return self.create_reference(result, object_id, None)

    def write_dataclass(
        self,
        obj: Container,
        object_id: int,
        subcon_class_name: str | None = None,
    ) -> dict[str, Any]:
        fields = sorted(
            dataclasses.fields(obj),
            key=lambda field: field.metadata.get("order", 0),
        )
        class_name = cast("str", inspect.getdoc(type(obj)))

        if (
            subcon_class_name is not None
            and class_name.endswith("List")
            and subcon_class_name != class_name
        ):
            class_name = f"{class_name}`1[[{subcon_class_name}, {LIBRARY_NAME_SINGING_TOOL_MODEL}]]"

        if class_name not in self.class_defs:
            result = {
                "obj": {
                    "class_info": {
                        "object_id": object_id,
                        "name": class_name,
                        "member_count": 0,
                        "member_names": [],
                    },
                    "member_type_info": {
                        "binary_type_enums": [],
                        "additional_infos": [],
                    },
                    "member_values": [],
                },
            }
            if class_name.startswith("System."):
                result["record_type_enum"] = RecordTypeEnum.SystemClassWithMembersAndTypes
            else:
                if class_name.startswith("SingingTool.Model."):
                    result["obj"]["library_id"] = self.model_library_id
                elif class_name.startswith("SingingTool.Library."):
                    result["obj"]["library_id"] = self.lib_library_id
                result["record_type_enum"] = RecordTypeEnum.ClassWithMembersAndTypes
            for field in fields:
                if field.metadata.get("alias"):
                    field_args = get_args(field.type)
                    field_origin = get_origin(field.type)
                    if field_origin in (list, XSBuf, XSBufList):
                        field_type = field_origin
                        if dataclasses.is_dataclass(field_args[0]):
                            subcon_class_name = inspect.getdoc(field_args[0])
                    elif (
                        field_origin == Union
                        and len(field_args) == 2
                        and issubclass(field_args[-1], type(None))
                    ):
                        field_type = field_args[0]
                    else:
                        field_type = field.type
                    if field.name == "edited_power_line" and self.svip_file["version"] != "7.0.0":
                        continue
                    if issubclass(field_type, str):
                        result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                            BinaryTypeEnum.String
                        )
                        result["obj"]["member_type_info"]["additional_infos"].append({"info": None})  # type: ignore[index]
                    elif issubclass(field_type, int | float | bool | enum.IntEnum):
                        result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                            BinaryTypeEnum.Primitive
                        )
                        if issubclass(field_type, bool):
                            primitive_type_enum = PrimitiveTypeEnum.Boolean
                        elif issubclass(field_type, float):
                            if field.name in ("pan", "sample_rate", "volume"):
                                primitive_type_enum = PrimitiveTypeEnum.Double
                            else:
                                primitive_type_enum = PrimitiveTypeEnum.Single
                        else:
                            primitive_type_enum = PrimitiveTypeEnum.Int32
                        result["obj"]["member_type_info"]["additional_infos"].append(  # type: ignore[index]
                            {"info": primitive_type_enum}
                        )
                    elif issubclass(field_type, list):
                        result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                            BinaryTypeEnum.Class
                        )
                        result["obj"]["member_type_info"]["additional_infos"].append(  # type: ignore[index]
                            {
                                "info": {
                                    "type_name": f"{subcon_class_name}[]",
                                    "library_id": self.model_library_id,
                                }
                            }
                        )
                    elif issubclass(field_type, bytes):
                        result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                            BinaryTypeEnum.PrimitiveArray
                        )
                        result["obj"]["member_type_info"]["additional_infos"].append(  # type: ignore[index]
                            {"info": PrimitiveTypeEnum.Byte}
                        )
                    elif dataclasses.is_dataclass(field_type):
                        sub_class_name = cast("str", inspect.getdoc(field_type))

                        if sub_class_name.endswith("List"):
                            sub_class_name = f"{sub_class_name}`1[[{subcon_class_name}, {LIBRARY_NAME_SINGING_TOOL_MODEL}]]"

                        if sub_class_name.startswith("System."):
                            result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                                BinaryTypeEnum.SystemClass
                            )
                            result["obj"]["member_type_info"]["additional_infos"].append(  # type: ignore[index]
                                {
                                    "info": sub_class_name,
                                }
                            )
                        else:
                            result["obj"]["member_type_info"]["binary_type_enums"].append(  # type: ignore[index]
                                BinaryTypeEnum.Class
                            )
                            result["obj"]["member_type_info"]["additional_infos"].append(  # type: ignore[index]
                                {
                                    "info": {
                                        "type_name": sub_class_name,
                                        "library_id": self.model_library_id
                                        if sub_class_name.startswith("SingingTool.Model.")
                                        else self.lib_library_id,
                                    },
                                }
                            )
                    else:
                        msg = f"Unknown type {field_type}"
                        raise TypeError(msg)
                    value = getattr(obj, field.name)
                    if value is None:
                        result["obj"]["member_values"].append(
                            {
                                "value": {
                                    "record_type_enum": RecordTypeEnum.ObjectNull,
                                    "obj": {},
                                }
                            }
                        )
                    elif issubclass(field_type, str):
                        result["obj"]["member_values"].append(
                            {"value": self.create_string(value or "")}
                        )
                    elif issubclass(field_type, int | float | bool | enum.IntEnum):
                        result["obj"]["member_values"].append({"value": value})
                    elif issubclass(field_type, list):
                        if len(value) > 0:
                            subcon_class_name = inspect.getdoc(type(value[0]))
                        result["obj"]["member_values"].append(
                            {
                                "value": self.write_binary_array(
                                    value,
                                    cast("str", subcon_class_name),
                                    self.model_library_id,
                                )
                            }
                        )
                    elif issubclass(field_type, bytes):
                        result["obj"]["member_values"].append(
                            {"value": self.create_primitive_array(getattr(obj, field.name))}
                        )
                    elif dataclasses.is_dataclass(field_type):
                        sub_class_name = cast("str", inspect.getdoc(field_type))

                        if sub_class_name.endswith("List"):
                            sub_class_name = f"{sub_class_name}`1[[{subcon_class_name}, {LIBRARY_NAME_SINGING_TOOL_MODEL}]]"
                        if field.name == "buf_1":
                            obj_id = self.id_max
                        elif hasattr(field_type, "value"):
                            self.id_max += 1
                            obj_id = self.id_max
                        else:
                            obj_id = self.enq()
                        if hasattr(field_type, "value"):
                            result["obj"]["member_values"].append(
                                {"value": self.write_dataclass(value, -obj_id, sub_class_name)}
                            )
                        elif field.name.endswith("_line") and not len(value.line_param):
                            result["obj"]["member_values"].append(
                                {
                                    "value": {
                                        "record_type_enum": RecordTypeEnum.ObjectNull,
                                        "obj": {},
                                    }
                                }
                            )
                        else:
                            result["obj"]["member_values"].append(
                                {"value": self.create_reference(value, obj_id, sub_class_name)}
                            )
                    else:
                        msg = f"Unknown type {field_type}"
                        raise TypeError(msg)
                    result["obj"]["class_info"]["member_names"].append(  # type: ignore[index]
                        field.metadata["alias"]
                    )
                    result["obj"]["class_info"]["member_count"] += 1  # type: ignore[index]
            self.class_defs[class_name] = result
            classes_by_id[self.cur_thread_id][object_id] = result["obj"]
        else:
            result = {
                "record_type_enum": RecordTypeEnum.ClassWithId,
                "obj": {
                    "object_id": object_id,
                    "metadata_id": self.class_defs[class_name]["obj"]["class_info"]["object_id"],
                    "member_values": [],
                },
            }
            for field in fields:
                if field.metadata.get("alias"):
                    field_args = get_args(field.type)
                    field_origin = get_origin(field.type)
                    if field_origin in (list, XSBuf, XSBufList):
                        field_type = field_origin
                        if dataclasses.is_dataclass(field_args[0]):
                            subcon_class_name = inspect.getdoc(field_args[0])
                    elif (
                        field_origin == Union
                        and len(field_args) == 2
                        and issubclass(field_args[-1], type(None))
                    ):
                        field_type = field_args[0]
                    else:
                        field_type = field.type

                    if field.name == "edited_power_line" and self.svip_file["version"] != "7.0.0":
                        continue
                    value = getattr(obj, field.name)
                    if value is None:
                        result["obj"]["member_values"].append(
                            {
                                "value": {
                                    "record_type_enum": RecordTypeEnum.ObjectNull,
                                    "obj": {},
                                }
                            }
                        )
                    elif issubclass(field_type, str):
                        result["obj"]["member_values"].append(
                            {"value": self.create_string(value or "")}
                        )
                    elif issubclass(field_type, int | float | bool | enum.IntEnum):
                        result["obj"]["member_values"].append({"value": value})
                    elif issubclass(field_type, list):
                        if len(value) > 0:
                            subcon_class_name = inspect.getdoc(type(value[0]))
                        result["obj"]["member_values"].append(
                            {
                                "value": self.write_binary_array(
                                    value,
                                    cast("str", subcon_class_name),
                                    self.model_library_id,
                                )
                            }
                        )
                    elif issubclass(field_type, bytes):
                        result["obj"]["member_values"].append(
                            {"value": self.create_primitive_array(getattr(obj, field.name))}
                        )
                    elif dataclasses.is_dataclass(field_type):
                        sub_class_name = cast("str", inspect.getdoc(field_type))

                        if sub_class_name.endswith("List"):
                            sub_class_name = f"{sub_class_name}`1[[{subcon_class_name}, {LIBRARY_NAME_SINGING_TOOL_MODEL}]]"
                        if field.name == "buf_1":
                            obj_id = self.id_max
                        elif hasattr(field_type, "value"):
                            self.id_max += 1
                            obj_id = self.id_max
                        else:
                            obj_id = self.enq()
                        if hasattr(field_type, "value"):
                            result["obj"]["member_values"].append(
                                {"value": self.write_dataclass(value, -obj_id, sub_class_name)}
                            )
                        elif field.name.endswith("_line") and not len(value.line_param):
                            result["obj"]["member_values"].append(
                                {
                                    "value": {
                                        "record_type_enum": RecordTypeEnum.ObjectNull,
                                        "obj": {},
                                    }
                                }
                            )
                        else:
                            result["obj"]["member_values"].append(
                                {"value": self.create_reference(value, obj_id, sub_class_name)}
                            )
                    else:
                        msg = f"Unknown type {field_type}"
                        raise TypeError(msg)
        return result



from PySide6.QtCore import QObject, Slot
from PySide6.QtQml import QmlElement

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from libresvip.utils.text import shorten_error_message

from .application import app

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


@QmlElement
class Clipboard(QObject):
    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent=parent)
        self.clipboard = app.clipboard()

    @Slot(str, result=bool)
    def set_clipboard(self, text: str) -> bool:
        self.clipboard.set_text(text)
        return True

    @Slot(str, result=str)
    def shorten_error_message(self, msg: str) -> str:
        return shorten_error_message(msg)



import pathlib
from typing import Annotated, Any, Literal

from pydantic import Field, ValidationInfo, model_validator
from typing_extensions import Self

from libresvip.model.base import BaseModel


class VOXFactoryNote(BaseModel):
    time: float
    midi: int
    name: str
    syllable: str
    ticks: float
    duration_ticks: float = Field(alias="durationTicks")
    duration: float = 0.25
    velocity: int = 1
    note_type: str | None = Field(None, alias="noteType")
    vibrato_depth: float | None = Field(None, alias="vibratoDepth")
    pre_bend: float | None = Field(None, alias="preBend")
    post_bend: float | None = Field(None, alias="postBend")
    harmonic_ratio: float | None = Field(None, alias="harmonicRatio")
    pitch_bends: list[float] = Field(default_factory=list, alias="pitchBends")


class VOXFactoryMetadata(BaseModel):
    style: str
    accent: str
    transpose: int
    harmonic_ratio: float = Field(alias="harmonicRatio")
    pitch_detection: str | None = Field(None, alias="pitchDetection")
    instrument: str | None = None


class VOXFactoryClipBase(BaseModel):
    name: str = ""
    start_quarter: float = Field(0, alias="startQuarter")
    offset_quarter: float = Field(0, alias="offsetQuarter")
    length: float
    use_source: bool = Field(True, alias="useSource")
    audio_data_key: str | None = Field(None, alias="audioDataKey")
    audio_data_order: list[str] = Field(default_factory=list, alias="audioDataOrder")
    audio_data_quarter: float = Field(0, alias="audioDataQuarter")
    note_bank: dict[str, VOXFactoryNote] = Field(default_factory=dict, alias="noteBank")
    note_order: list[str] = Field(default_factory=list, alias="noteOrder")
    next_note_index: int = Field(0, alias="nextNoteIndex")
    pinned_audio_data_order: list[str] = Field(default_factory=list, alias="pinnedAudioDataOrder")
    metadata: VOXFactoryMetadata | None = None


class VOXFactoryVocalClip(VOXFactoryClipBase):
    type: Literal["vocal"] = "vocal"
    length_type: Literal["quarter"] = Field("quarter", alias="lengthType")


class VOXFactoryAudioClip(VOXFactoryClipBase):
    type: Literal["audio"] = "audio"
    length_type: Literal["time"] = Field("time", alias="lengthType")
    source_audio_data_key: str = Field(alias="sourceAudioDataKey")

    @model_validator(mode="after")
    def extract_audio(self, info: ValidationInfo) -> Self:
        if (
            info.context is not None
            and info.context["extract_audio"]
            and not hasattr(info.context["path"], "protocol")
        ):
            archive_audio_path = f"resources/{self.source_audio_data_key}"
            if not (
                audio_path := (info.context["path"].parent / self.name).with_suffix(
                    pathlib.Path(archive_audio_path).suffix
                )
            ).exists():
                audio_path.write_bytes(info.context["archive_file"].read(archive_audio_path))
        return self


class VOXFactoryAudioViewProperty(BaseModel):
    view: str = "waveform"
    colormap: str = "Heated Metal"
    window: str = "Hann"
    window_size: int = Field(1024, alias="windowSize")
    hop_size: int = Field(256, alias="hopSize")
    f_min: float = Field(27.5, alias="fMin")
    f_max: float | None = Field(None, alias="fMax")
    level_min: float | None = Field(None, alias="levelMin")
    level_max: float | None = Field(None, alias="levelMax")
    level_scale: str = Field("dB", alias="levelScale")
    num_bins: int = Field(230, alias="numBins")
    bins_per_octave: int = Field(24, alias="binsPerOctave")


class VOXFactoryDevice(BaseModel):
    type: str
    track_type: str = Field(alias="trackType")
    name: str
    data: dict[str, Any]
    on: bool


class VOXFactoryTrackBase(BaseModel):
    name: str = ""
    instrument: str | None = None
    h: int = 3
    color: str = "#7878f1"
    volume: float = 1.0
    pan: float = 0.0
    solo: bool = False
    mute: bool = False
    arm: bool = False
    clip_order: list[str] = Field(alias="clipOrder")
    device_bank: dict[str, VOXFactoryDevice] = Field(default_factory=dict, alias="deviceBank")
    device_order: list[str] = Field(default_factory=list, alias="deviceOrder")
    audio_view_property: VOXFactoryAudioViewProperty = Field(
        default_factory=VOXFactoryAudioViewProperty, alias="audioViewProperty"
    )


class VOXFactoryVocalTrack(VOXFactoryTrackBase):
    type: Literal["vocal"] = "vocal"
    clip_bank: dict[str, VOXFactoryVocalClip] = Field(alias="clipBank")


class VOXFactoryAudioTrack(VOXFactoryTrackBase):
    type: Literal["audio"] = "audio"
    clip_bank: dict[str, VOXFactoryAudioClip] = Field(alias="clipBank")


VOXFactoryTrack = Annotated[
    VOXFactoryVocalTrack | VOXFactoryAudioTrack,
    Field(discriminator="type"),
]


class VOXFactorySelectedClipBankItem(BaseModel):
    track_key: str = Field(alias="trackKey")
    clip_key: str = Field(alias="clipKey")


class VOXFactorySelectedNoteBankItem(BaseModel):
    track_key: str = Field(alias="trackKey")
    clip_key: str = Field(alias="clipKey")
    note_key: str = Field(alias="noteKey")


class VOXFactoryAudioData(BaseModel):
    sample_rate: int = Field(alias="sampleRate")
    number_of_channels: int = Field(alias="numberOfChannels")
    sample_length: int = Field(alias="sampleLength")
    metadata: VOXFactoryMetadata | None = None


class VOXFactoryProject(BaseModel):
    version: str = "0.12.0"
    tempo: float
    time_signature: list[int] = Field(alias="timeSignature")
    project_name: str = Field("Untitled Project", alias="projectName")
    track_bank: dict[str, VOXFactoryTrack] = Field(alias="trackBank")
    track_order: list[str] = Field(default_factory=list, alias="trackOrder")
    selected_track_bank: list[str] = Field(default_factory=list, alias="selectedTrackBank")
    selected_clip_bank: list[VOXFactorySelectedClipBankItem] = Field(
        default_factory=list, alias="selectedClipBank"
    )
    selected_note_bank: list[VOXFactorySelectedNoteBankItem] = Field(
        default_factory=list, alias="selectedNoteBank"
    )
    audio_data_bank: dict[str, VOXFactoryAudioData] = Field(
        default_factory=dict, alias="audioDataBank"
    )



import dataclasses
import functools

import portion

from libresvip.core.constants import MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS
from libresvip.core.exceptions import NotesOverlappedError
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note
from libresvip.model.portamento import PortamentoPitch
from libresvip.utils.translation import gettext_lazy as _


@dataclasses.dataclass
class PitchSimulator:
    synchronizer: TimeSynchronizer
    portamento: PortamentoPitch
    note_list: dataclasses.InitVar[list[Note]]
    interval_dict: PiecewiseIntervalDict = dataclasses.field(default_factory=PiecewiseIntervalDict)

    def __post_init__(self, note_list: list[Note]) -> None:
        if not note_list:
            return
        current_note = note_list[0]
        max_portamento_percent = self.portamento.max_inter_time_percent
        if self.portamento.vocaloid_mode:
            max_portamento_ticks = max_portamento_percent * current_note.length
            if max_portamento_ticks >= 60:
                max_portamento_ticks = 60
            elif current_note.length <= 120:
                max_portamento_ticks = current_note.length / 2
            max_portamento_time = self.synchronizer.get_duration_secs_from_ticks(
                int(current_note.end_pos - max_portamento_ticks * 1.4),
                int(current_note.end_pos - max_portamento_ticks * 0.4),
            )
        else:
            max_portamento_time = self.portamento.max_inter_time_in_secs

        current_head = self.synchronizer.get_actual_secs_from_ticks(current_note.start_pos)
        current_dur = self.synchronizer.get_duration_secs_from_ticks(
            current_note.start_pos, current_note.end_pos
        )
        current_portamento = min(current_dur * max_portamento_percent, max_portamento_time)

        self.interval_dict[portion.closedopen(0.0, current_head)] = current_note.key_number
        prev_portamento_end = current_head
        for next_note in note_list[1:]:
            if current_note.end_pos > next_note.start_pos:
                msg = _("Notes Overlapped")
                raise NotesOverlappedError(msg)
            elif (
                self.portamento.vocaloid_mode
                and next_note.start_pos - current_note.end_pos
                >= MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS
            ):
                max_portamento_time = 0
            next_head = self.synchronizer.get_actual_secs_from_ticks(next_note.start_pos)
            next_dur = self.synchronizer.get_duration_secs_from_ticks(
                next_note.start_pos, next_note.end_pos
            )
            next_portamento = min(next_dur * max_portamento_percent, max_portamento_time)
            if self.portamento.vocaloid_mode:
                middle_pos = (
                    current_note.end_pos
                    if next_note.lyric == "-"
                    else (next_note.start_pos + current_note.end_pos) / 2
                ) - max_portamento_ticks * 0.4  # pyright: ignore[reportPossiblyUnboundVariable]
                interval = self.synchronizer.get_duration_secs_from_ticks(
                    int(middle_pos - max_portamento_ticks),  # pyright: ignore[reportPossiblyUnboundVariable]
                    int(middle_pos),
                )
                middle_time = self.synchronizer.get_actual_secs_from_ticks(int(middle_pos))
            else:
                interval = (next_head - current_head - current_dur) / 2
                middle_time = (next_head + current_head + current_dur) / 2
            if interval <= max_portamento_time:
                current_portamento_start = middle_time - current_portamento
                current_portamento_end = middle_time + next_portamento
            else:
                current_portamento_start = middle_time - max_portamento_time
                current_portamento_end = middle_time + max_portamento_time
            self.interval_dict[
                portion.closedopen(prev_portamento_end, current_portamento_start)
            ] = current_note.key_number
            if current_note.key_number == next_note.key_number:
                self.interval_dict[
                    portion.closedopen(current_portamento_start, current_portamento_end)
                ] = current_note.key_number
            elif current_portamento_start < current_portamento_end:
                self.interval_dict[
                    portion.closedopen(current_portamento_start, current_portamento_end)
                ] = functools.partial(  # type: ignore[call-arg]
                    self.portamento.inter_func,
                    start=(current_portamento_start, current_note.key_number),
                    end=(current_portamento_end, next_note.key_number),
                )
            current_note = next_note
            current_head = next_head
            current_dur = next_dur
            current_portamento = next_portamento
            prev_portamento_end = current_portamento_end
            if self.portamento.vocaloid_mode:
                max_portamento_ticks = max_portamento_percent * current_note.length
                if max_portamento_ticks >= 60:
                    max_portamento_ticks = 60
                elif current_note.length <= 120:
                    max_portamento_ticks = current_note.length / 2
                max_portamento_time = self.synchronizer.get_duration_secs_from_ticks(
                    int(middle_pos - max_portamento_ticks * 1.4),  # pyright: ignore[reportPossiblyUnboundVariable]
                    int(middle_pos - max_portamento_ticks * 0.4),  # pyright: ignore[reportPossiblyUnboundVariable]
                )
        self.interval_dict[portion.closedopen(prev_portamento_end, portion.inf)] = (
            current_note.key_number
        )

    def pitch_at_ticks(self, ticks: int) -> float | None:
        return self.pitch_at_secs(self.synchronizer.get_actual_secs_from_ticks(ticks))

    def pitch_at_secs(self, secs: float) -> float | None:
        if value := self.interval_dict.get(secs):
            return value * 100



import dataclasses
import math
from collections.abc import MutableSequence
from typing import cast

from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info

from .model import (
    PitchPoint,
    UCurve,
    UNote,
    UPitch,
    URendererSettings,
    USTXProject,
    UTempo,
    UTimeSignature,
    UTrack,
    UVibrato,
    UVoicePart,
    UWavePart,
)
from .options import OpenUtauEnglishPhonemizerCompatibility, OutputOptions
from .util import BasePitchGenerator
from .utils import lyric_util


@dataclasses.dataclass
class UstxGenerator:
    options: OutputOptions

    def generate_project(self, os_project: Project) -> USTXProject:
        # 节拍
        ustx_time_signatures = [
            self.generate_time_signature(ts) for ts in os_project.time_signature_list
        ]
        if not ustx_time_signatures:
            ustx_time_signatures.append(UTimeSignature())
        first_bar_length = int(
            1920 * ustx_time_signatures[0].beat_per_bar / ustx_time_signatures[0].beat_unit
        )

        # 曲速
        tempos = [self.generate_tempo(x, first_bar_length) for x in os_project.song_tempo_list]
        if not tempos:
            tempos.append(UTempo())

        ustx_project = USTXProject(
            ustx_version="0.6",
            tempos=tempos,
            bpm=tempos[0].bpm,
            time_signatures=ustx_time_signatures,
            voice_parts=[],
            wave_parts=[],
        )

        ustx_project.tracks = []
        for track_no, os_track in enumerate(os_project.track_list):
            ustx_project.tracks.append(self.generate_track(os_track))
            if isinstance(os_track, SingingTrack):  # 合成音轨
                ustx_project.tracks[-1].singer = os_track.ai_singer_name
                ustx_project.voice_parts.append(
                    self.generate_voice_part(os_track, track_no, ustx_project, first_bar_length)
                )
            elif wav_part := self.generate_wave_part(os_track, track_no):  # 伴奏音轨
                ustx_project.wave_parts.append(wav_part)
        return ustx_project

    @staticmethod
    def generate_tempo(os_tempo: SongTempo, first_bar_length: int = 1920) -> UTempo:
        return UTempo(
            position=max(os_tempo.position - first_bar_length, 0),
            bpm=os_tempo.bpm,
        )

    @staticmethod
    def generate_time_signature(
        os_time_signature: TimeSignature,
    ) -> UTimeSignature:
        return UTimeSignature(
            bar_position=os_time_signature.bar_index,
            beat_per_bar=os_time_signature.numerator,
            beat_unit=os_time_signature.denominator,
        )

    @staticmethod
    def generate_track(os_track: Track) -> UTrack:
        return UTrack(
            track_name=os_track.title,
            singer="",
            phonemizer="OpenUtau.Core.DefaultPhonemizer",  # 默认音素器
            renderer_settings=URendererSettings(renderer="CLASSIC"),
            mute=os_track.mute,
            solo=os_track.solo,
            volume=math.log10(os_track.volume) * 10,  # 绝对音量转对数音量
        )

    def generate_voice_part(
        self,
        os_track: SingingTrack,
        track_no: int,
        ustx_project: USTXProject,
        first_bar_length: int,
    ) -> UVoicePart:
        ustx_voice_part = UVoicePart(name=os_track.title, track_no=track_no, position=0, notes=[])
        if not os_track.note_list:
            return ustx_voice_part
        # 转换音符
        ustx_voice_part.notes.extend(self.generate_notes(os_track.note_list))
        # 转换音高曲线
        self.generate_pitch(
            ustx_voice_part,
            ustx_project,
            cast(
                "MutableSequence[tuple[int, int]]",
                os_track.edited_params.pitch.points.root,
            ),
            first_bar_length,
        )

        return ustx_voice_part

    @staticmethod
    def generate_wave_part(os_track: InstrumentalTrack, track_no: int) -> UWavePart | None:
        if track_info := audio_track_info(os_track.audio_file_path):
            return UWavePart(
                name=os_track.title,
                track_no=track_no,
                position=os_track.offset,
                relative_path=os_track.audio_file_path,
                file_duration_ms=track_info.duration,
            )

    def generate_notes(self, os_notes: list[Note]) -> list[UNote]:
        notes: list[UNote] = []
        last_note_end_pos = -480  # 上一个音符的结束时间
        last_note_key_number = 60  # 上一个音符的音高
        last_syllable_index = 1
        for i, os_note in enumerate(os_notes):
            snap_first = last_note_end_pos >= os_note.start_pos
            y0 = (last_note_key_number - os_note.key_number) * 10 if snap_first else 0
            lyric = lyric_util.get_symbol_removed_lyric(os_note.lyric)  # 去除标点符号
            if os_note.pronunciation:  # 如果有发音, 则用发音
                lyric = os_note.pronunciation
            if lyric == "-":  # OpenUTAU中的连音符为+
                lyric = "+"
            elif lyric == "+":
                if (
                    self.options.english_phonemizer_compatibility
                    == OpenUtauEnglishPhonemizerCompatibility.ARPA
                ):
                    last_syllable_index += 1
                    lyric = f"+{last_syllable_index}"
                else:
                    for j in range(i - 1, 0, -1):
                        if os_notes[j].lyric == "-":
                            notes[j].lyric = "+~"
                        else:
                            break
            elif (
                self.options.english_phonemizer_compatibility
                == OpenUtauEnglishPhonemizerCompatibility.ARPA
            ):
                last_syllable_index = 1
            if len(lyric) == 2 and lyric_util.is_punctuation(lyric[1]):  # 删除标点符号
                lyric = lyric[:1]
            notes.append(
                UNote(
                    position=os_note.start_pos,
                    duration=os_note.length,
                    tone=os_note.key_number,
                    lyric=lyric,
                    pitch=UPitch(
                        snap_first=snap_first,
                        data=[
                            PitchPoint(x=-40, y=y0, shape="io"),
                            PitchPoint(x=0, y=0, shape="io"),
                        ],
                    ),
                    vibrato=UVibrato(
                        length=0,
                        period=175,
                        depth=25,
                        in_value=10,
                        out=10,
                        shift=0,
                        drift=0,
                    ),
                )
            )
            last_note_end_pos = os_note.end_pos
            last_note_key_number = os_note.key_number
        return notes

    @staticmethod
    def generate_pitch(
        part: UVoicePart,
        project: USTXProject,
        os_pitch: MutableSequence[tuple[int, int]],
        first_bar_length: int,
    ) -> None:
        pitch_start = BasePitchGenerator.pitch_start
        pitch_interval = BasePitchGenerator.pitch_interval
        base_pitch = BasePitchGenerator(project).base_pitch(part)  # 生成基础音高线
        pitch_end_x = len(base_pitch) * pitch_interval + first_bar_length + 1

        # 如果os_pitch为空
        if not os_pitch:
            os_pitch = [(0, -1), (pitch_end_x, -1)]
        # 如果os_pitch末尾不完整
        if os_pitch[-1][0] < pitch_end_x:
            os_pitch.append((os_pitch[-1][0], -1))
            os_pitch.append((pitch_end_x, -1))

        os_pitch_pointer = 0  # 当前位于输入os_pitch的第几个采样点
        pitd = UCurve(abbr="pitd", xs=[], ys=[])
        for i in range(len(base_pitch)):
            time = (
                i * pitch_interval + pitch_start
            )  # 当前openutau采样点的时间, 以tick为单位, 从0开始
            while os_pitch[os_pitch_pointer + 1][0] <= time + first_bar_length:
                os_pitch_pointer += 1
            # 此时, os_pitch_pointer对应的位置恰好在time之前(或等于), 区间[os_pitch_pointer,os_pitch_pointer+1)包含time
            if os_pitch[os_pitch_pointer][1] < 0:  # 间断点
                pitd.xs.append(time)
                pitd.ys.append(0)
            elif os_pitch[os_pitch_pointer][0] == os_pitch[os_pitch_pointer + 1][0]:
                pitd.xs.append(time)
                pitd.ys.append(
                    round(
                        (os_pitch[os_pitch_pointer][1] + os_pitch[os_pitch_pointer + 1][1]) / 2
                        - int(base_pitch[i])
                    )
                )
            else:  # 有实际曲线存在
                x1 = os_pitch[os_pitch_pointer][0] - first_bar_length
                x2 = os_pitch[os_pitch_pointer + 1][0] - first_bar_length
                y1 = os_pitch[os_pitch_pointer][1]
                y2 = os_pitch[os_pitch_pointer + 1][1]
                pitd.xs.append(time)
                pitd.ys.append(
                    round((y2 - y1) * (time - x1) / (x2 - x1) + y1 - int(base_pitch[i]))
                )  # 线性插值
        if not pitd.is_empty:
            part.curves.append(pitd)



from importlib.resources import files

from bidict import bidict

from libresvip.core.compat import json

singer_data_path = files("libresvip.plugins.svip3") / "singers.json"
singers_data = bidict(json.loads(singer_data_path.read_text(encoding="utf-8")))



CS_DBLCLKS = 8
GWL_STYLE = -16
GWL_EXSTYLE = -20
HTCAPTION = 2
HTSYSMENU = 3
HTMINBUTTON = 8
HTMAXBUTTON = 9
HTLEFT = 10
HTRIGHT = 11
HTTOP = 12
HTTOPLEFT = 13
HTTOPRIGHT = 14
HTBOTTOM = 15
HTBOTTOMLEFT = 16
HTBOTTOMRIGHT = 17
HTBORDER = 18
HTCLOSE = 20
SC_MAXIMIZE = 61488
SC_RESTORE = 61728
SIZE_MINIMIZED = 1
TPM_LEFTBUTTON = 0
TPM_RIGHTBUTTON = 2
TPM_LEFTALIGN = 0
TPM_CENTERALIGN = 4
TPM_RIGHTALIGN = 8
TPM_TOPALIGN = 0
TPM_VCENTERALIGN = 16
TPM_BOTTOMALIGN = 32
TPM_HORIZONTAL = 0
TPM_VERTICAL = 64
TPM_NONOTIFY = 128
TPM_RETURNCMD = 256
TPM_RECURSE = 1
WS_EX_LAYERED = 0x00080000
WS_MAXIMIZEBOX = 65536
WS_CAPTION = 12582912
WS_THICKFRAME = 262144
WM_NCCALCSIZE = 131
WM_NCHITTEST = 132
WM_NCMOUSEMOVE = 160
WM_NCLBUTTONDOWN = 161
WM_NCLBUTTONUP = 162
WM_NCLBUTTONDBLCLK = 163
WM_NCMOUSEHOVER = 672
WM_NCMOUSELEAVE = 674
WM_SIZE = 5
WM_SYSCOMMAND = 274
WM_ACTIVATE = 6
WVR_HREDRAW = 256
WVR_VREDRAW = 512
WVR_REDRAW = WVR_HREDRAW | WVR_VREDRAW



from __future__ import annotations

import itertools
import math
import statistics
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    NamedTuple,
    cast,
)

from more_itertools import batched, minmax
from pydantic import (
    Field,
    FieldSerializationInfo,
    RootModel,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from libresvip.core.time_interval import RangeInterval
from libresvip.model.base import BaseModel
from libresvip.model.point import PointList
from libresvip.utils.audio import audio_path_validator
from libresvip.utils.music_math import HermiteInterpolator

from .enums import AcepLyricsLanguage
from .singers import DEFAULT_SEED, DEFAULT_SINGER, DEFAULT_SINGER_ID

if TYPE_CHECKING:
    from collections.abc import Callable


class AcepAnchorPoint(NamedTuple):
    pos: float
    value: float


class AcepAnchorPoints(PointList[AcepAnchorPoint], RootModel[list[AcepAnchorPoint]]):
    root: list[AcepAnchorPoint] = Field(default_factory=list)


class AcepParamCurve(BaseModel):
    curve_type: str = Field("data", alias="type")
    offset: int = 0
    values: list[float] = Field(default_factory=list)
    points: AcepAnchorPoints | None = None
    points_vuv: list[float] | None = Field(None, alias="pointsVUV")

    @field_validator("points", mode="before")
    @classmethod
    def validate_points(cls, points: list[float], _info: ValidationInfo) -> AcepAnchorPoints:
        return AcepAnchorPoints(
            root=[AcepAnchorPoint._make(each) for each in batched(points or [], 2)]
        )

    @field_serializer("points", when_used="json-unless-none")
    def serialize_points(
        self, points: AcepAnchorPoints, _info: FieldSerializationInfo
    ) -> list[float]:
        return list(chain.from_iterable(points.root))

    @model_validator(mode="after")
    def points2values(self) -> Self:
        if self.curve_type == "anchor" and self.points is not None and len(self.points.root):
            interpolator = HermiteInterpolator(
                cast("list[tuple[float, float]]", self.points.root),
            )
            self.offset = math.floor(self.points.root[0].pos)
            self.values = interpolator.interpolate(
                list(
                    range(
                        self.offset,
                        math.ceil(self.points.root[-1].pos) + 1,
                    )
                ),
            )
        return self

    def transform(self, value_transform: Callable[[float], float]) -> AcepParamCurve:
        return self.model_copy(
            update={"values": [value_transform(each) for each in self.values]},
            deep=True,
        )


class AcepParamCurveList(RootModel[list[AcepParamCurve]]):
    root: list[AcepParamCurve] = Field(default_factory=list)

    def plus(
        self,
        others: AcepParamCurveList | None,
        default_value: float,
        transform: Callable[[float], float],
    ) -> AcepParamCurveList:
        if not others:
            return self
        result_ranges = RangeInterval(
            [
                (curve.offset, curve.offset + len(curve.values))
                for curve in (self.root + others.root)
            ],
        ).sub_ranges()
        result_curve_list = type(self)()
        for start, end in result_ranges:
            result_curve = AcepParamCurve()
            result_curve.offset = start
            result_curve.values = [0.0] * (end - start)
            for self_curve in (curve for curve in self.root if start <= curve.offset < end):
                index = self_curve.offset - start
                for value in self_curve.values:
                    result_curve.values[index] = value
                    index += 1
            for other_curve in (curve for curve in others.root if start <= curve.offset < end):
                index = other_curve.offset - start
                for value in other_curve.values:
                    if result_curve.values[index] == 0.0:
                        result_curve.values[index] = default_value
                    result_curve.values[index] += transform(value)
                    index += 1
            result_curve_list.root.append(result_curve)
        return result_curve_list

    def exclude(self, predicate: Callable[[float], bool]) -> AcepParamCurveList:
        result = type(self)()
        for curve in self.root:
            buffer: list[float] = []
            pos = curve.offset
            for value in curve.values:
                pos += 1
                if predicate(value):
                    if buffer:
                        result.root.append(AcepParamCurve(offset=pos - len(buffer), values=buffer))
                        buffer.clear()
                else:
                    buffer.append(value)

            if buffer:
                result.root.append(AcepParamCurve(offset=pos - len(buffer), values=buffer))

        return result

    def z_score_normalize(self, d: float = 1, b: float = 0) -> AcepParamCurveList:
        if not self.root:
            return self
        points = [*itertools.chain.from_iterable(curve.values for curve in self.root)]
        miu = statistics.mean(points)
        sigma = statistics.stdev(points)
        return type(self)(
            root=[curve.transform(lambda x: (x - miu) / sigma * d + b) for curve in self.root]
        )

    def minmax_normalize(self, r: float = 1, b: float = 0) -> AcepParamCurveList:
        if not self.root:
            return self
        min_, max_ = minmax(
            itertools.chain.from_iterable(curve.values for curve in self.root),
            default=(0, 0),
        )
        return type(self)(
            root=[
                curve.transform(lambda x: r * (2 * (x - min_) / (max_ - min_) - 1) + b)
                for curve in self.root
            ]
            if abs(max_ - min_) > 1e-3
            else [curve.transform(lambda x: 0) for curve in self.root]
        )


class AcepMaster(BaseModel):
    gain: float = Field(default=0.0, le=6.0)


class AcepTempo(BaseModel):
    bpm: float = 0.0
    position: int = 0
    is_lerp: bool | None = Field(False, alias="isLerp")


class AcepParams(BaseModel):
    pitch_delta: AcepParamCurveList = Field(default_factory=AcepParamCurveList, alias="pitchDelta")
    energy: AcepParamCurveList = Field(default_factory=AcepParamCurveList)
    breathiness: AcepParamCurveList = Field(default_factory=AcepParamCurveList)
    tension: AcepParamCurveList = Field(default_factory=AcepParamCurveList)
    falsetto: AcepParamCurveList = Field(default_factory=AcepParamCurveList)
    gender: AcepParamCurveList = Field(default_factory=AcepParamCurveList)
    real_energy: AcepParamCurveList = Field(default_factory=AcepParamCurveList, alias="realEnergy")
    real_breathiness: AcepParamCurveList = Field(
        default_factory=AcepParamCurveList, alias="realBreathiness"
    )
    real_tension: AcepParamCurveList = Field(
        default_factory=AcepParamCurveList, alias="realTension"
    )
    real_falsetto: AcepParamCurveList = Field(
        default_factory=AcepParamCurveList, alias="realFalsetto"
    )
    vuv: AcepParamCurveList | None = Field(default_factory=AcepParamCurveList)


class AcepVibrato(BaseModel):
    start: float = 0.0
    amplitude: float = 0.0
    frequency: float = 0.0
    attack_len: float = Field(0.0, alias="attackLen")
    release_len: float = Field(0.0, alias="releaseLen")
    release_vol: float = Field(0.0, alias="releaseVol")
    phase: float = 0.0
    start_pos: float = Field(0.0, alias="startPos")
    release_level: float = Field(0.0, alias="releaseLevel")
    release_ratio: float = Field(0.0, alias="releaseRatio")
    attack_level: float = Field(0.0, alias="attackLevel")
    attack_ratio: float = Field(0.0, alias="attackRatio")


class AcepNote(BaseModel):
    pos: int = 0
    dur: int = 0
    pitch: int = 0
    language: AcepLyricsLanguage = AcepLyricsLanguage.CHINESE
    lyric: str = ""
    pronunciation: str | None = None
    freezed_default_syllable: str | None = Field(None, alias="freezedDefaultSyllable")
    new_line: bool = Field(False, alias="newLine")
    consonant_len: int | None = Field(None, alias="consonantLen")
    head_consonants: list[float] | None = Field(default_factory=list, alias="headConsonants")
    tail_consonants: list[float] | None = Field(default_factory=list, alias="tailConsonants")
    syllable: str = ""
    br_len: float = Field(0.0, alias="brLen")
    vibrato: AcepVibrato | None = None
    extra_info: dict[str, Any] = Field(default_factory=dict, alias="extraInfo")


class AcepPattern(BaseModel):
    name: str = ""
    pos: float = 0.0
    dur: float = 0.0
    clip_pos: float = Field(0.0, alias="clipPos")
    clip_dur: float = Field(0.0, alias="clipDur")
    enabled: bool | None = True
    color: str | None = None
    extra_info: dict[str, Any] = Field(default_factory=dict, alias="extraInfo")


class AcepAnalysedBeat(BaseModel):
    beat_times: list[float] = Field(default_factory=list, alias="beatTimes")
    length: float = 0
    offset: int = 0
    scales: list[int] = Field(default_factory=list)


class AcepAudioFadeShape(BaseModel):
    offset_x: float = Field(0.0, alias="offsetX")
    offset_y: float = Field(0.0, alias="offsetY")


class AcepAudioFadeEffect(BaseModel):
    length: float = 0.0
    crossfade: bool | None = False
    shape: AcepAudioFadeShape = Field(default_factory=AcepAudioFadeShape)


class AcepAudioPattern(AcepPattern):
    path: str = ""
    gain: float | None = None
    analysed_beat: AcepAnalysedBeat | None = Field(None, alias="analysedBeat")
    time_unit: str | None = Field("sec", alias="timeUnit")
    fade_in: AcepAudioFadeEffect | None = Field(None, alias="fadeIn")
    fade_out: AcepAudioFadeEffect | None = Field(None, alias="fadeOut")
    validate_path = field_validator("path", mode="before")(audio_path_validator)


class AcepVocalPattern(AcepPattern):
    language: AcepLyricsLanguage = AcepLyricsLanguage.CHINESE
    extend_lyrics: str = Field("", alias="extendLyrics")
    notes: list[AcepNote] = Field(default_factory=list)
    time_unit: str | None = Field("tick", alias="timeUnit")
    parameters: AcepParams = Field(default_factory=AcepParams)


class AcepTrackProperties(BaseModel):
    name: str = ""
    color: str = "#91bcdc"
    gain: float = Field(0.0, le=6.0)
    pan: float = Field(0.0, le=10.0, ge=-10.0)
    mute: bool = False
    solo: bool = False
    record: bool = False
    channel: int | None = 0
    listen: bool | None = False
    extra_info: dict[str, Any] = Field(default_factory=dict, alias="extraInfo")
    built_in_fx: dict[str, Any] = Field(default_factory=dict, alias="builtInFx")


class AcepEmptyTrack(AcepTrackProperties, BaseModel):
    type_: Literal["empty"] = Field(default="empty", alias="type")


class AcepAudioTrack(AcepTrackProperties, BaseModel):
    type_: Literal["audio"] = Field(default="audio", alias="type")
    patterns: list[AcepAudioPattern] = Field(default_factory=list)


class AcepSeedComposition(BaseModel):
    code: int = DEFAULT_SEED
    lock: bool = True
    style: float = 1.0
    timbre: float = 1.0


class AcepCustomSinger(BaseModel):
    composition: list[AcepSeedComposition] = Field(default_factory=list)
    state: str = "Unmixed"
    name: str = DEFAULT_SINGER
    singer_id: int | None = Field(DEFAULT_SINGER_ID, alias="id")
    head: int | None = -1
    router: int | None = 1
    group: str | None = ""


class AcepSingerConfig(BaseModel):
    singer: AcepCustomSinger = Field(default_factory=AcepCustomSinger)
    gain: float = 0.0
    mute: bool = False
    random_seed: int = Field(0, alias="randomSeed")


class AcepVocalTrack(AcepTrackProperties, BaseModel):
    type_: Literal["sing"] = Field(default="sing", alias="type")
    singer: int | AcepCustomSinger | None = None
    language: AcepLyricsLanguage = AcepLyricsLanguage.CHINESE
    patterns: list[AcepVocalPattern] = Field(default_factory=list)
    choir_info: dict[str, Any] = Field(default_factory=dict, alias="choirInfo")
    singers: list[AcepSingerConfig] = Field(default_factory=list)

    @model_validator(mode="after")
    def migrate_singer_attr(self) -> Self:
        if self.singer is not None:
            if isinstance(self.singer, int):
                self.singers.append(
                    AcepSingerConfig(singer=AcepCustomSinger(singer_id=self.singer))
                )
            else:
                self.singers.append(AcepSingerConfig(singer=self.singer))
        return self

    def __len__(self) -> int:
        if not len(self.patterns):
            return 0
        last_pattern = self.patterns[-1]
        return int(last_pattern.pos + last_pattern.clip_dur - last_pattern.clip_pos)


class AcepChord(BaseModel):
    addeds: list[Literal["7", "j7", "b9", "9", "#9", "11", "#11", "b13", "13"]] = Field(
        default_factory=list
    )
    bass: int
    dur: int = 0
    root: int = -1
    type_: Literal["", "maj", "min", "dim", "sus2", "sus4", "aug"] = Field(default="", alias="type")


class AcepChordPattern(AcepPattern):
    chords: list[AcepChord] = Field(default_factory=list)
    time_unit: str | None = Field("tick", alias="timeUnit")


class AcepChordTrack(AcepTrackProperties, BaseModel):
    type_: Literal["chord"] = Field(default="chord", alias="type")
    patterns: list[AcepChordPattern] = Field(default_factory=list)


AcepTrack = Annotated[
    AcepAudioTrack | AcepEmptyTrack | AcepVocalTrack | AcepChordTrack,
    Field(discriminator="type_"),
]


class AcepTimeSignature(BaseModel):
    bar_pos: int = Field(0, alias="barPos")
    numerator: int = 4
    denominator: int = 4


class AcepProject(BaseModel):
    beats_per_bar: int = Field(4, alias="beatsPerBar")
    color_index: int = Field(0, alias="colorIndex")
    pattern_individual_color_index: int | None = Field(0, alias="patternIndividualColorIndex")
    debug_info: dict[str, Any] = Field(default_factory=dict, alias="debugInfo")
    duration: int = 0
    extra_info: dict[str, Any] = Field(default_factory=dict, alias="extraInfo")
    master: AcepMaster = Field(default_factory=AcepMaster)
    piano_cells: int = Field(2147483646, alias="pianoCells")
    tempos: list[AcepTempo] = Field(default_factory=list)
    track_cells: int = Field(2147483646, alias="trackCells")
    tracks: list[AcepTrack] = Field(default_factory=list)
    loop: bool | None = False
    loop_start: int | None = Field(0, alias="loopStart")
    loop_end: int | None = Field(7680, alias="loopEnd")
    version: int = 9
    version_revision: int | None = Field(0, alias="versionRevision")
    merged_pattern_index: int = Field(0, alias="mergedPatternIndex")
    record_pattern_index: int = Field(0, alias="recordPatternIndex")
    singer_library_id: str | None = "1200593006"
    time_signatures: list[AcepTimeSignature] = Field(default_factory=list, alias="timeSignatures")
    track_control_panel_w: int | None = Field(0, alias="trackControlPanelW")

    @model_validator(mode="after")
    def migrate_time_signatures(self) -> Self:
        if not self.time_signatures:
            self.time_signatures.append(
                AcepTimeSignature(
                    bar_pos=0,
                    numerator=self.beats_per_bar,
                    denominator=4,
                )
            )
        return self



import dataclasses
import math

import pypinyin

from libresvip.core.exceptions import NoTrackError
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.translation import gettext_lazy as _

from .model import NNInfoLine, NNNote, NNPoints, NNProject, NNTimeSignature
from .options import OutputOptions


@dataclasses.dataclass
class NiaoniaoGenerator:
    options: OutputOptions
    length_multiplier: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> NNProject:
        self.length_multiplier = 60 if self.options.version == 19 else 30
        if self.options.track_index < 0:
            if (
                first_singing_track := next(
                    (track for track in project.track_list if isinstance(track, SingingTrack)),
                    None,
                )
            ) is None:
                msg = _("No singing track found")
                raise NoTrackError(msg)
        else:
            first_singing_track = project.track_list[self.options.track_index]
        nn_time_signature = self.generate_time_signature(project.time_signature_list)
        nn_tempo = self.generate_tempo(project.song_tempo_list)
        nn_info_line = NNInfoLine(
            version=self.options.version,
            time_signature=nn_time_signature,
            tempo=nn_tempo,
        )
        nn_project = NNProject(
            info_line=nn_info_line,
            notes=self.generate_notes(first_singing_track),
        )
        nn_project.note_count = len(nn_project.notes)
        if nn_project.note_count:
            nn_info_line.bar_count = math.ceil(
                (nn_project.notes[-1].start + nn_project.notes[-1].duration)
                * self.length_multiplier
                / self.first_bar_length
            )
        return nn_project

    def generate_time_signature(self, time_signature_list: list[TimeSignature]) -> NNTimeSignature:
        if not len(time_signature_list):
            self.first_bar_length = 1920
            return NNTimeSignature()
        self.first_bar_length = int(time_signature_list[0].bar_length())
        return NNTimeSignature(
            numerator=time_signature_list[0].numerator,
            denominator=time_signature_list[0].denominator,
        )

    def generate_tempo(self, tempo_list: list[SongTempo]) -> float:
        return tempo_list[0].bpm

    def generate_notes(self, singing_track: SingingTrack) -> list[NNNote]:
        nn_notes = []
        for note in singing_track.note_list:
            nn_note = NNNote(
                lyric=note.lyric,
                pronunciation=note.pronunciation or " ".join(pypinyin.lazy_pinyin(note.lyric)),
                key=88 - note.key_number,
                start=note.start_pos // self.length_multiplier,
                duration=note.length // self.length_multiplier,
            )
            if singing_track.edited_params.pitch:
                nn_note.pitch = self.generate_pitch(singing_track.edited_params.pitch, note)
            nn_notes.append(nn_note)
        return nn_notes

    def generate_pitch(self, pitch_param_curve: ParamCurve, note: Note) -> NNPoints:
        sample_time_list = [
            note.start_pos + self.first_bar_length + int((note.length / 100.0) * i)
            for i in range(100)
        ]
        pitch_param_in_note = [
            p
            for p in pitch_param_curve.points.root
            if p.x >= note.start_pos + self.first_bar_length
            and p.x < note.end_pos + self.first_bar_length
        ]

        pitch_param_time_in_note = dict(pitch_param_in_note)

        nn_pitch_param = []
        for sample_time in sample_time_list:
            if (pitch := pitch_param_time_in_note.get(sample_time)) is None:
                distance = -1
                value = 50

                for point in pitch_param_in_note:
                    if distance > abs(point.x - sample_time) or distance == -1:
                        distance = abs(point.x - sample_time)
                        value = 50 + (
                            0 if point.y == -100 else round((point.y - note.key_number * 100) / 12)
                        )

                nn_pitch_param.append(value)

            elif pitch == -100:
                nn_pitch_param.append(50)
            else:
                nn_pitch_param.append(50 + round((pitch - note.key_number * 100) / 12))
        buffer = []
        previous_node = nn_pitch_param[0]
        previous_node_index = 0
        for i in range(len(nn_pitch_param)):
            if nn_pitch_param[i] == previous_node:
                buffer.append(nn_pitch_param[i])
            else:
                for j in range(len(buffer)):
                    nn_pitch_param[previous_node_index + j] = round(
                        previous_node + j * (nn_pitch_param[i] - buffer[j]) / len(buffer)
                    )
                buffer.clear()

            if nn_pitch_param[i] != previous_node:
                previous_node_index = i
                previous_node = nn_pitch_param[i]

        return NNPoints(points=nn_pitch_param)



import dataclasses
import re
from typing import cast

import more_itertools

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    ParamCurve,
    Params,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import hz2midi, note2midi

from .options import InputOptions
from .utils.models.ds_file import DsItem, DsProject

CENTS_RE = re.compile(r"[+-]\d+$")


@dataclasses.dataclass
class DiffSingerParser:
    options: InputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def parse_project(self, ds_project: DsProject) -> Project:
        song_tempo_list = [SongTempo(position=0, bpm=self.options.tempo)]
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=[TimeSignature(bar_index=0, numerator=4, denominator=4)],
            track_list=[
                SingingTrack(
                    note_list=self.parse_notes(ds_project.root),
                    edited_params=Params(pitch=self.parse_pitch(ds_project.root)),
                )
            ],
        )

    def parse_notes(self, ds_items: list[DsItem]) -> list[Note]:
        all_notes = []
        for ds_item in ds_items:
            notes: list[Note] = []
            cur_time = self.synchronizer.get_actual_ticks_from_secs(float(ds_item.offset))
            prev_is_breath = False
            for lyric_index, slur_group in enumerate(
                more_itertools.split_before(
                    enumerate(ds_item.note_slur or []),
                    lambda pair: pair[1] == 0,
                )
            ):
                if not ds_item.note_dur:
                    break
                for note_index, is_slur in slur_group:
                    text = ds_item.text[lyric_index]
                    note_dur = ds_item.note_dur[note_index]
                    note = ds_item.note_seq[note_index]
                    next_time = self.synchronizer.get_actual_ticks_from_secs_offset(
                        int(cur_time), note_dur
                    )
                    if text == "SP":
                        pass
                    elif text == "AP":
                        prev_is_breath = True
                    else:
                        midi_key = note2midi(CENTS_RE.sub("", note))
                        if not is_slur:
                            notes.append(
                                Note(
                                    start_pos=int(cur_time),
                                    length=int(next_time) - int(cur_time),
                                    key_number=midi_key,
                                    lyric=text,
                                    head_tag="V" if prev_is_breath else None,
                                )
                            )
                            prev_is_breath = False
                        else:
                            notes.append(
                                Note(
                                    start_pos=int(cur_time),
                                    length=int(next_time) - int(cur_time),
                                    key_number=midi_key,
                                    lyric="-",
                                )
                            )
                    cur_time = next_time
            all_notes.extend(notes)
        return all_notes

    def parse_pitch(self, ds_items: list[DsItem]) -> ParamCurve:
        points = Points(root=[])
        points.append(Point.start_point())
        if self.options.import_pitch:
            for ds_item in ds_items:
                if ds_item.f0_timestep is None or ds_item.f0_seq is None:
                    continue
                f0_timestep = ds_item.f0_timestep
                points.append(
                    Point(
                        round(self.synchronizer.get_actual_ticks_from_secs(float(ds_item.offset)))
                        + 1920,
                        -100,
                    )
                )
                points.extend(
                    [
                        Point(
                            round(
                                self.synchronizer.get_actual_ticks_from_secs(
                                    cast("float", ds_item.offset) + f0_timestep * i
                                )
                            )
                            + 1920,
                            round(hz2midi(float(f0)) * 100),
                        )
                        for i, f0 in enumerate(cast("list[float]", ds_item.f0_seq))
                    ]
                )
                points.append(
                    Point(
                        round(
                            self.synchronizer.get_actual_ticks_from_secs(
                                cast("float", ds_item.offset)
                                + f0_timestep * (len(cast("list[float]", ds_item.f0_seq)) - 1)
                            )
                        )
                        + 1920,
                        -100,
                    )
                )
        points.append(Point.end_point())
        return ParamCurve(points=points)



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
)
from libresvip.utils.translation import gettext_lazy as _

from .model import VocaloidLanguage


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    is_ai_singer: bool = Field(
        False,
        title=_("Is AI Singer"),
        description=_("The default voicebank is AI singer or not"),
    )
    default_lang_id: VocaloidLanguage = Field(
        VocaloidLanguage.SIMPLIFIED_CHINESE,
        title=_("Default language"),
        description=_("Default language id of voicebank and notes"),
    )
    default_comp_id: str = Field(
        "BL8CEAM5N4XN3LFK",
        title=_("Default Comp ID"),
        description=_("Default comp_id of voicebank"),
    )
    default_singer_name: str = Field(
        "Luo_Tianyi_Wan",
        title=_("Default Singer Name"),
        description=_("Default singer name of voicebank"),
    )



from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class BaseTask:
    name: str | None = ""
    path: str | None = ""
    stem: str | None = ""
    ext: str | None = ""
    tmp_path: str | None = ""

    running: bool | None = False
    success: bool | None = None
    error: str | None = ""
    warning: str | None = ""
    child_tasks: list[BaseTask] = dataclasses.field(default_factory=list)



import dataclasses
import pathlib

from libresvip.core.constants import (
    DEFAULT_CHINESE_LYRIC,
    DEFAULT_ENGLISH_LYRIC,
    DEFAULT_JAPANESE_LYRIC,
    DEFAULT_KOREAN_LYRIC,
)
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.translation import gettext_lazy as _

from .constants import (
    BPM_RATE,
    PITCH_BEND_NAME,
    PITCH_BEND_SENSITIVITY_NAME,
)
from .model import (
    VocaloidAudioTrack,
    VocaloidLanguage,
    VocaloidNotes,
    VocaloidPartPitchData,
    VocaloidPoint,
    VocaloidProject,
    VocaloidTimeSig,
    VocaloidTracks,
)
from .options import InputOptions
from .vocaloid_pitch import pitch_from_vocaloid_parts


@dataclasses.dataclass
class VocaloidParser:
    options: InputOptions
    path: pathlib.Path
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    comp_id2name: dict[str, str] = dataclasses.field(init=False)

    def parse_project(self, vpr_project: VocaloidProject) -> Project:
        self.comp_id2name = {
            voice.comp_id: voice.name
            for voice in vpr_project.voices
            if voice.comp_id is not None and voice.name is not None
        }
        song_tempo_list = self.parse_tempos(vpr_project.master_track.tempo.events)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            time_signature_list=self.parse_time_signatures(
                vpr_project.master_track.time_sig.events
            ),
            song_tempo_list=song_tempo_list,
            track_list=self.parse_tracks(vpr_project.tracks),
        )

    def parse_time_signatures(
        self, vocaloid_time_signatures: list[VocaloidTimeSig]
    ) -> list[TimeSignature]:
        time_signatures = [
            TimeSignature(
                bar_index=time_signature.bar,
                numerator=time_signature.numer,
                denominator=time_signature.denom,
            )
            for time_signature in vocaloid_time_signatures
        ]
        self.first_bar_length = round(time_signatures[0].bar_length())
        return time_signatures

    def parse_tempos(self, tempos: list[VocaloidPoint]) -> list[SongTempo]:
        return [
            SongTempo(
                position=tempo.pos,
                bpm=tempo.value / BPM_RATE,
            )
            for tempo in tempos
            if isinstance(tempo.value, int | float)
        ]

    def parse_tracks(self, tracks: list[VocaloidTracks]) -> list[Track]:
        track_list = []
        for track in tracks:
            if isinstance(track, VocaloidAudioTrack):
                if self.options.import_instrumental_track:
                    for part in track.parts:
                        if part.wav is None:
                            continue
                        instrumental_track = InstrumentalTrack(
                            title=part.name,
                            offset=part.pos,
                            mute=track.is_muted,
                            solo=track.is_solo_mode,
                            audio_file_path=str(
                                self.path.parent / (part.wav.original_name or part.wav.name)
                            ),
                        )
                        track_list.append(instrumental_track)
            else:
                for part in track.parts:
                    comp_id = ""
                    supported_lang_ids = []
                    if part.voice is not None:
                        comp_id = part.voice.comp_id or ""
                        supported_lang_ids.append(part.voice.lang_id)
                    elif part.ai_voice is not None:
                        comp_id = part.ai_voice.comp_id or ""
                        supported_lang_ids.extend(part.ai_voice.lang_ids)
                    if supported_lang_ids:
                        main_lang_id = supported_lang_ids[0]
                    else:
                        main_lang_id = VocaloidLanguage.SIMPLIFIED_CHINESE
                    if main_lang_id == VocaloidLanguage.JAPANESE:
                        default_lyric = DEFAULT_JAPANESE_LYRIC
                    elif main_lang_id == VocaloidLanguage.KOREAN:
                        default_lyric = DEFAULT_KOREAN_LYRIC
                    elif main_lang_id == VocaloidLanguage.SIMPLIFIED_CHINESE:
                        default_lyric = DEFAULT_CHINESE_LYRIC
                    else:
                        default_lyric = DEFAULT_ENGLISH_LYRIC
                    notes, direct_pitch_points = self.parse_notes(
                        part.notes, part.pos, default_lyric
                    )
                    singing_track = SingingTrack(
                        title=part.name,
                        mute=track.is_muted,
                        solo=track.is_solo_mode,
                        note_list=notes,
                        ai_singer_name=self.comp_id2name.get(comp_id, ""),
                    )
                    if self.options.import_pitch:
                        if direct_pitch_points:
                            singing_track.edited_params.pitch.points.root.extend(
                                direct_pitch_points
                            )
                        elif (
                            part_data := VocaloidPartPitchData(
                                start_pos=part.pos,
                                pit=part.get_controller_events(PITCH_BEND_NAME),
                                pbs=part.get_controller_events(PITCH_BEND_SENSITIVITY_NAME),
                            )
                        ) and (
                            part_pitch := pitch_from_vocaloid_parts(
                                [part_data],
                                self.synchronizer,
                                singing_track.note_list,
                                self.first_bar_length,
                            )
                        ) is not None:
                            singing_track.edited_params.pitch = part_pitch
                    track_list.append(singing_track)
        return track_list

    def parse_notes(
        self, notes: list[VocaloidNotes], pos: int, default_lyric: str
    ) -> tuple[list[Note], list[Point]]:
        note_list: list[Note] = []
        pitch_points: list[Point] = []
        if len(notes):
            next_pos = None
            for note in notes[::-1]:
                normalized_duration = note.duration or 0
                if self.options.import_pitch and note.direct_pitches is not None:
                    note_pitch_points: list[Point] = []
                    for pitch in note.direct_pitches:
                        if pitch.pos < 0 or (note.duration and pitch.pos > note.duration):
                            continue
                        if isinstance(pitch.value, int | float):
                            if not note_pitch_points or note_pitch_points[-1].y == -100:
                                note_pitch_points.append(
                                    Point(
                                        x=note.pos + pos + pitch.pos + self.first_bar_length, y=-100
                                    )
                                )
                            note_pitch_points.append(
                                Point(
                                    x=note.pos + pos + pitch.pos + self.first_bar_length,
                                    y=int(pitch.value) + note.number * 100,
                                )
                            )
                        else:
                            note_pitch_points.extend(
                                [
                                    Point(
                                        x=note.pos + pos + pitch.pos + self.first_bar_length,
                                        y=note.number * 100,
                                    ),
                                    Point(
                                        x=note.pos + pos + pitch.pos + self.first_bar_length, y=-100
                                    ),
                                ]
                            )
                    pitch_points = note_pitch_points + pitch_points
                if next_pos is not None:
                    distance = next_pos - note.pos
                    if distance < normalized_duration:
                        normalized_duration = distance
                        if normalized_duration > 0:
                            show_warning(_("Note overlap detected, cutting note ") + note.lyric)
                if normalized_duration > 0:
                    note_list.insert(
                        0,
                        Note(
                            start_pos=note.pos + pos,
                            length=normalized_duration,
                            key_number=note.number,
                            lyric=note.lyric or default_lyric,
                            pronunciation=None,
                        ),
                    )
                else:
                    show_warning(_("Note overlap detected, skipping note ") + note.lyric)
                next_pos = note.pos
        if pitch_points:
            pitch_points.insert(0, Point.start_point())
            pitch_points.append(Point.end_point())
        return note_list, pitch_points



import sys

from PySide6.QtGui import QIcon, QPixmap

from libresvip.core.constants import res_dir
from libresvip.gui.modules import (
    LocaleSwitcher,
    app,
    app_close_event,
    event_loop,
    qml_engine,
)

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401


def startup() -> None:
    qml_engine.load(":/qml/main.qml")
    if not qml_engine.root_objects():
        sys.exit(-1)
    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())


def run() -> None:
    icon_pixmap = QPixmap()
    icon_pixmap.load_from_data((res_dir / "libresvip.ico").read_bytes())
    app.application_name = "LibreSVIP"
    app.organization_name = "org.soulmelody.libresvip"
    app.window_icon = QIcon(icon_pixmap)
    locale_switcher = LocaleSwitcher()
    locale_switcher.translator_initialized.connect(startup)
    locale_switcher.switch_language(locale_switcher.get_language())


if __name__ == "__main__":
    run()



import ast
from typing import Any, Literal

from parsimonious import Grammar, NodeVisitor
from parsimonious.nodes import Node
from pydantic import BaseModel, Field

from .constants import MAX_ACCEPTED_BPM

ust_grammar = Grammar(
    r"""
    ust_project =
        ust_header?
        ust_setting_section
        ust_track*
        newline*

    ust_header =
        "[#VERSION]"
        newline "UST Version" "="? float
        (newline "Charset" "=" value)?

    ust_setting_section =
        newline? "[#SETTING]"
        (newline ust_setting_line)*

    ust_time_signature =
        "(" int "/" int  "/" int ")"

    ust_setting_line =
        ("UstVersion" "=" float) /
        ("Tempo" "=" value) /
        ("TimeSignatures" "=" ust_time_signature ("," ust_time_signature)* ","?) /
        ("Tracks" "=" int) /
        (~"Project(Name)?" "=" value) /
        ("VoiceDir" "=" value) /
        ("OutFile" "=" value) /
        ("CacheDir" "=" value) /
        ("Tool1" "=" value) /
        ("Tool2" "=" value) /
        ("Mode2" "=" bool) /
        ("Autoren" "=" bool) /
        ("MapFirst" "=" bool) /
        ("Flags" "=" value)

    ust_track =
        ust_note*
        (newline ust_track_end)?

    ust_note =
        newline ust_note_head
        (newline ust_note_attr " "*)*

    ust_note_head =
        "[#" ~"(\\d+|PREV|NEXT|INSERT|DELETE)" "]"

    ust_track_end = "[#TRACKEND]"

    ust_tag_entry = ust_note_head / ust_track_end

    utau_pitch_bend_mode =
        "s" / "r" / "j" / "null" / ""

    ust_pitch_bend_type =
        "5" / "OldData"

    ust_note_attr =
        ("Length" "=" float) /
        ("Duration" "=" float) /
        ("Lyric" "=" value) /
        ("NoteNum" "=" int) /
        ("Delta" "=" int) /
        ("PreUtterance" "=" value) /
        ("VoiceOverlap" "=" float) /
        ("Intensity" "=" float) /
        (~"Modulation|Moduration" "=" float) /
        ("StartPoint" "=" float) /
        ("Envelope" "=" ust_envelope) /
        ("Tempo" "=" value) /
        ("Velocity" "=" float) /
        ("Label" "=" value) /
        ("Flags" "=" value) /
        ("PBType" "=" ust_pitch_bend_type) /
        ("PBStart" "=" float) /
        (~"Piches|Pitches|PitchBend" "=" int ("," int)* ) /
        ("PBS" "=" optional_float (~";|," optional_float)*) /
        ("PBW" "=" optional_float ("," optional_float)*) /
        ("PBY" "=" optional_float ("," optional_float)*) /
        ("VBR" "=" optional_float ("," optional_float)*) /
        ("PBM" "=" utau_pitch_bend_mode ("," utau_pitch_bend_mode)* ) /
        ("stptrim" "=" float) /
        ("layer" "=" int) /
        ("@preuttr" "=" float) /
        ("@overlap" "=" float) /
        ("@stpoint" "=" float) /
        ("@filename" "=" value) /
        ("@alias" "=" value) /
        ("@cache" "=" value) /
        (~"\\$?[^=\r\n]+" "=" value) /
        (!ust_tag_entry value)

    ust_envelope =
        float ("," float){6}
        (
            (",%," float? ("," float? ("," float?)?)?) /
            (",," float) /
            ",%" /
            ("," float)*
        )

    value = ~"[^\r\n]*"
    newline = ~"\r?\n"
    bool = "1" / "0" / "True" / "False"
    optional_float = float / "null" / ""
    float = ((int "."? digit*) / ("-"? "." digits)) (~"[eE][+-]?" digits)?
    int = "-"? digits
    digits = digit+
    digit = ~"[0-9]"
    """
)

UTAUPitchBendMode = Literal["s", "r", "j", ""]

UTAUPitchBendType = Literal["OldData", "5"]

OptionalFloat = float | Literal[""]


class UtauNoteVibrato(BaseModel):
    length: float  # percentage of the note's length
    period: float  # milliSec
    depth: float = 0  # cent
    fade_in: float = 0  # percentage of the vibrato's length
    fade_out: float = 0  # percentage of the vibrato's length
    phase_shift: float = 0  # percentage of period
    shift: float = 0  # percentage of depth


class UTAUEnvelope(BaseModel):
    p1: float
    p2: float
    p3: float
    v1: float
    v2: float
    v3: float
    v4: float
    p4: float | None = None
    p5: float | None = None
    v5: float | None = None
    other_points: list[float] | None = None


class UTAUNote(BaseModel):
    note_type: str
    length: int
    lyric: str
    note_num: int
    duration: int | None = None
    delta: int | None = None
    pre_utterance: str | None = None
    label: str | None = None
    flags: str | None = None
    envelope: UTAUEnvelope | None = None
    voice_overlap: float | None = None
    intensity: float | None = None
    modulation: float | None = None
    start_point: float | None = None
    tempo: float | None = None
    velocity: float | None = None
    pitchbend_start: float | None = None
    pitchbend_type: UTAUPitchBendType | None = None
    pitch_bend_points: list[int] = Field(default_factory=list)
    pbs: list[OptionalFloat] = Field(default_factory=list)
    pbw: list[OptionalFloat] = Field(default_factory=list)
    pby: list[OptionalFloat] = Field(default_factory=list)
    pbm: list[UTAUPitchBendMode] = Field(default_factory=list)
    vbr: UtauNoteVibrato | None = None
    stp_trim: float | None = None
    layer: int | None = None
    at_preutterance: float | None = None
    at_overlap: float | None = None
    at_start_point: float | None = None
    sample_filename: str | None = None
    alias: str | None = None
    cache_location: str | None = None


class UTAUTrack(BaseModel):
    notes: list[UTAUNote] = Field(default_factory=list)


class UTAUTimeSignature(BaseModel):
    numerator: int
    denominator: int
    bar_index: int


class UTAUProject(BaseModel):
    ust_version: float | None = 1.2
    project_name: str = "New Project"
    track_count: int = 1
    pitch_mode2: bool = False
    charset: str | None = None
    tempo: float | None = None
    voice_dir: str | None = None
    out_file: str | None = None
    cache_dir: str | None = None
    tool1: str | None = None
    tool2: str | None = None
    flags: str | None = None
    autoren: bool | None = None
    map_first: bool | None = None
    time_signatures: list[UTAUTimeSignature] = Field(default_factory=list)
    track: list[UTAUTrack] = Field(default_factory=list)


class UstVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.pending_metadata: dict[str, Any] = {}
        self.pending_note_attrs: dict[str, Any] = {}
        self.pending_notes: list[UTAUNote] = []

    @staticmethod
    def tempo2bpm(tempo: str) -> float:
        return float(tempo.replace(",", "."))

    def visit_ust_project(self, node: Node, visited_children: list[Any]) -> UTAUProject:
        self.pending_metadata["track"] = [
            track for track in visited_children[2] if track is not None
        ]
        return UTAUProject(**self.pending_metadata)

    def visit_ust_header(self, node: Node, visited_children: list[Any]) -> None:
        self.pending_metadata["ust_version"] = visited_children[4]
        if isinstance(visited_children[-1], list):
            self.pending_metadata["charset"] = visited_children[-1][-1][-1]

    def visit_ust_setting_line(self, node: Node, visited_children: list[Any]) -> None:
        key = visited_children[0][0].text
        match key:
            case "Tempo":
                if (tempo_value := self.tempo2bpm(visited_children[0][2])) < MAX_ACCEPTED_BPM:
                    self.pending_metadata["tempo"] = tempo_value
            case "UstVersion":
                self.pending_metadata["ust_version"] = visited_children[0][2]
            case "Tracks":
                self.pending_metadata["track_count"] = visited_children[0][2]
            case "Project" | "ProjectName":
                self.pending_metadata["project_name"] = visited_children[0][2]
            case "Mode2":
                self.pending_metadata["pitch_mode2"] = visited_children[0][2]
            case "VoiceDir":
                self.pending_metadata["voice_dir"] = visited_children[0][2]
            case "OutFile":
                self.pending_metadata["out_file"] = visited_children[0][2]
            case "CacheDir":
                self.pending_metadata["cache_dir"] = visited_children[0][2]
            case "Tool1":
                self.pending_metadata["tool1"] = visited_children[0][2]
            case "Tool2":
                self.pending_metadata["tool2"] = visited_children[0][2]
            case "Autoren":
                self.pending_metadata["autoren"] = visited_children[0][2]
            case "MapFirst":
                self.pending_metadata["map_first"] = visited_children[0][2]
            case "Flags":
                self.pending_metadata["flags"] = visited_children[0][2]
            case "TimeSignatures":
                numerator, denominator, bar_index = visited_children[0][2][1::2]
                self.pending_metadata["time_signatures"] = [
                    UTAUTimeSignature(
                        numerator=numerator,
                        denominator=denominator,
                        bar_index=bar_index,
                    )
                ]
                for pair in visited_children[0][3]:
                    numerator, denominator, bar_index = pair[1][1::2]
                    self.pending_metadata["time_signatures"].append(
                        UTAUTimeSignature(
                            numerator=numerator,
                            denominator=denominator,
                            bar_index=bar_index,
                        )
                    )

    def visit_ust_envelope(self, node: Node, visited_children: list[Any]) -> UTAUEnvelope:
        kwargs = {"p1": visited_children[0]}
        (
            kwargs["p2"],
            kwargs["p3"],
            kwargs["v1"],
            kwargs["v2"],
            kwargs["v3"],
            kwargs["v4"],
        ) = (pair[1] for pair in visited_children[1][:6])
        if isinstance(visited_children[2][0], list):
            if isinstance(visited_children[2][0][0], list):
                kwargs["other_points"] = [pair[1] for pair in visited_children[2][0]]
            elif visited_children[2][0][0].text == ",,":
                kwargs["p4"] = visited_children[2][0][1]
            elif visited_children[2][0][0].text == ",%,":
                if isinstance(visited_children[2][0][1], list):
                    kwargs["p4"] = visited_children[2][0][1][0]
                if isinstance(visited_children[2][0][2], list):
                    if isinstance(visited_children[2][0][2][0][1], list):
                        kwargs["p5"] = visited_children[2][0][2][0][1][0]
                    if isinstance(visited_children[2][0][2][0][2], list) and isinstance(
                        visited_children[2][0][2][0][2][0][1], list
                    ):
                        kwargs["v5"] = visited_children[2][0][2][0][2][0][1][0]
        return UTAUEnvelope(**kwargs)

    def visit_value(self, node: Node, visited_children: list[Any]) -> str:
        return node.text

    def visit_utau_pitch_bend_mode(self, node: Node, visited_children: list[Any]) -> str:
        return "" if node.text == "null" else node.text

    def visit_ust_pitch_bend_type(self, node: Node, visited_children: list[Any]) -> str:
        return node.text

    def visit_ust_note_attr(self, node: Node, visited_children: list[Any]) -> None:
        key = visited_children[0][0].text
        match key:
            case "Length":
                self.pending_note_attrs["length"] = visited_children[0][2]
            case "Lyric":
                self.pending_note_attrs["lyric"] = visited_children[0][2]
            case "NoteNum":
                self.pending_note_attrs["note_num"] = visited_children[0][2]
            case "Tempo":
                self.pending_note_attrs["tempo"] = self.tempo2bpm(visited_children[0][2])
            case "PBType":
                self.pending_note_attrs["pitchbend_type"] = visited_children[0][2]
            case "PBStart":
                self.pending_note_attrs["pitchbend_start"] = visited_children[0][2]
            case "Modulation" | "Moduration":
                self.pending_note_attrs["modulation"] = visited_children[0][2]
            case "Envelope":
                self.pending_note_attrs["envelope"] = visited_children[0][2]
            case "Flags":
                self.pending_note_attrs["flags"] = visited_children[0][2]
            case "Intensity":
                self.pending_note_attrs["intensity"] = visited_children[0][2]
            case "Velocity":
                self.pending_note_attrs["velocity"] = visited_children[0][2]
            case "PreUtterance":
                self.pending_note_attrs["pre_utterance"] = visited_children[0][2]
            case "StartPoint":
                self.pending_note_attrs["start_point"] = visited_children[0][2]
            case "Delta":
                self.pending_note_attrs["delta"] = visited_children[0][2]
            case "Duration":
                self.pending_note_attrs["duration"] = visited_children[0][2]
            case "VoiceOverlap":
                self.pending_note_attrs["voice_overlap"] = visited_children[0][2]
            case "Label":
                self.pending_note_attrs["label"] = visited_children[0][2]
            case "Piches" | "Pitches" | "PitchBend":
                self.pending_note_attrs["pitch_bend_points"] = [visited_children[0][2]]
                if isinstance(visited_children[0][3], list):
                    self.pending_note_attrs["pitch_bend_points"].extend(
                        [pair[1] for pair in visited_children[0][3]]
                    )
            case "PBS" | "PBM" | "PBW" | "PBY":
                self.pending_note_attrs[key.lower()] = [visited_children[0][2]]
                if isinstance(visited_children[0][3], list):
                    self.pending_note_attrs[key.lower()].extend(
                        [pair[1] for pair in visited_children[0][3]]
                    )
            case "VBR":
                vibrato_kwargs = {"length": visited_children[0][2]}
                for i, pair in enumerate(visited_children[0][3]):
                    if isinstance(pair[1], float):
                        if i == 0:
                            vibrato_kwargs["period"] = pair[1]
                        elif i == 1:
                            vibrato_kwargs["depth"] = pair[1]
                        elif i == 2:
                            vibrato_kwargs["fade_in"] = pair[1]
                        elif i == 3:
                            vibrato_kwargs["fade_out"] = pair[1]
                        elif i == 4:
                            vibrato_kwargs["phase_shift"] = pair[1]
                        elif i == 5:
                            vibrato_kwargs["shift"] = pair[1]
                self.pending_note_attrs["vbr"] = UtauNoteVibrato(**vibrato_kwargs)
            case "stptrim":
                self.pending_note_attrs["stp_trim"] = visited_children[0][2]
            case "layer":
                self.pending_note_attrs["layer"] = visited_children[0][2]
            case "@preuttr":
                self.pending_note_attrs["at_preutterance"] = visited_children[0][2]
            case "@overlap":
                self.pending_note_attrs["at_overlap"] = visited_children[0][2]
            case "@stpoint":
                self.pending_note_attrs["at_start_point"] = visited_children[0][2]
            case "@filename":
                self.pending_note_attrs["sample_filename"] = visited_children[0][2]
            case "@alias":
                self.pending_note_attrs["alias"] = visited_children[0][2]
            case "@cache":
                self.pending_note_attrs["cache_location"] = visited_children[0][2]
            # case _: # ignored

    def visit_ust_note_head(self, node: Node, visited_children: list[Any]) -> None:
        if len(self.pending_note_attrs):
            self.pending_notes.append(UTAUNote(**self.pending_note_attrs))
            self.pending_note_attrs = {}
        self.pending_note_attrs["note_type"] = visited_children[1].text

    def visit_ust_track(self, node: Node, visited_children: list[Any]) -> UTAUTrack | None:
        if len(self.pending_note_attrs):
            self.pending_notes.append(UTAUNote(**self.pending_note_attrs))
            self.pending_note_attrs = {}
        if len(self.pending_notes):
            ust_track = UTAUTrack(notes=self.pending_notes)
            self.pending_notes = []
            return ust_track

    def visit_bool(self, node: Node, visited_children: list[Any]) -> bool:
        return ast.literal_eval(node.text)

    def visit_int(self, node: Node, visited_children: list[Any]) -> int:
        return int(node.text)

    def visit_float(self, node: Node, visited_children: list[Any]) -> float:
        return float(node.text)

    def visit_optional_float(self, node: Node, visited_children: list[Any]) -> OptionalFloat:
        return (
            visited_children[0]
            if len(visited_children) == 1 and isinstance(visited_children[0], float)
            else ""
        )

    def generic_visit(self, node: Node, visited_children: list[Any]) -> Any:
        return visited_children or node
